from __future__ import annotations

import typing
from dataclasses import dataclass

from packaging.version import Version, parse

from pylav.helpers.singleton import SingletonCachedByKey
from pylav.storage.database.cache.decodators import maybe_cached
from pylav.storage.database.cache.model import CachedModel
from pylav.storage.database.tables.version import BotVersionRow


@dataclass(eq=True, slots=True, unsafe_hash=True, order=True, kw_only=True, frozen=True)
class BotVersion(CachedModel, metaclass=SingletonCachedByKey):
    """A class to represent the version of the bot in the database"""

    id: int

    def get_cache_key(self) -> str:
        """Get the cache key for the bot version."""
        return f"{self.id}"

    @maybe_cached
    async def fetch_version(self) -> Version:
        """Fetch the version of the bot from the database"""
        data = (
            await BotVersionRow.select(BotVersionRow.version)
            .where(BotVersionRow.bot == self.id)
            .first()
            .output(load_json=True, nested=True)
        )
        return typing.cast(Version, parse(data["version"] if data else BotVersionRow.version.default))

    async def update_version(self, version: Version | str) -> None:
        """Update the version of the bot in the database"""
        await BotVersionRow.insert(BotVersionRow(bot=self.id, version=str(version))).on_conflict(
            action="DO UPDATE", target=BotVersionRow.bot, values=[BotVersionRow.version]
        )



from __future__ import annotations

import typing
from dataclasses import dataclass

import discord
from piccolo.querystring import QueryString

from pylav.compat import json
from pylav.helpers.misc import TimedFeature
from pylav.helpers.singleton import SingletonCachedByKey
from pylav.storage.database.cache.decodators import maybe_cached
from pylav.storage.database.cache.model import CachedModel
from pylav.storage.database.tables.players import PlayerRow
from pylav.type_hints.dict_typing import JSON_DICT_TYPE


@dataclass(eq=True, slots=True, unsafe_hash=True, order=True, kw_only=True, frozen=True)
class PlayerConfig(CachedModel, metaclass=SingletonCachedByKey):
    id: int
    bot: int

    def get_cache_key(self) -> str:
        return f"{self.id}:{self.bot}"

    @classmethod
    async def create_global(cls, bot: int) -> None:
        """Create the player in the database"""
        data = {
            "enabled": False,
            "time": 60,
        }
        await PlayerRow.insert(
            PlayerRow(
                id=0,
                bot=bot,
                volume=1000,
                max_volume=1000,
                shuffle=True,
                auto_shuffle=True,
                auto_play=True,
                self_deaf=True,
                empty_queue_dc=data,
                alone_dc=data,
                alone_pause=data,
            )
        ).on_conflict(action="DO NOTHING", target=(PlayerRow.id, PlayerRow.bot))

    async def delete(self) -> None:
        """Delete the player from the database"""
        await PlayerRow.delete().where((PlayerRow.id == self.id) & (PlayerRow.bot == self.bot))
        await self.invalidate_cache()

    @maybe_cached
    async def fetch_all(self) -> JSON_DICT_TYPE:
        """Get all players from the database"""
        data = (
            await PlayerRow.select()
            .where((PlayerRow.id == self.id) & (PlayerRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )
        if data:
            del data["primary_key"]
            data["empty_queue_dc"] = TimedFeature.from_dict(data["empty_queue_dc"])
            data["alone_dc"] = TimedFeature.from_dict(data["alone_dc"])
            data["alone_pause"] = TimedFeature.from_dict(data["alone_pause"])
            data["extras"] = data["extras"]
            data["effects"] = data["effects"]
            return data
        return {
            "id": self.id,
            "bot": self.bot,
            "volume": PlayerRow.volume.default,
            "max_volume": PlayerRow.max_volume.default,
            "auto_play_playlist_id": PlayerRow.auto_play_playlist_id.default,
            "text_channel_id": PlayerRow.text_channel_id.default,
            "notify_channel_id": PlayerRow.notify_channel_id.default,
            "forced_channel_id": PlayerRow.forced_channel_id.default,
            "repeat_current": PlayerRow.repeat_current.default,
            "repeat_queue": PlayerRow.repeat_queue.default,
            "shuffle": PlayerRow.shuffle.default,
            "auto_shuffle": PlayerRow.auto_shuffle.default,
            "auto_play": PlayerRow.auto_play.default,
            "self_deaf": PlayerRow.self_deaf.default,
            "empty_queue_dc": TimedFeature.from_dict(json.loads(PlayerRow.empty_queue_dc.default)),
            "alone_dc": TimedFeature.from_dict(json.loads(PlayerRow.alone_dc.default)),
            "alone_pause": TimedFeature.from_dict(json.loads(PlayerRow.alone_pause.default)),
            "extras": json.loads(PlayerRow.extras.default),
            "effects": json.loads(PlayerRow.effects.default),
            "dj_users": PlayerRow.dj_users.default,
            "dj_roles": PlayerRow.dj_roles.default,
        }

    @maybe_cached
    async def exists(self) -> bool:
        """Check if the player exists in the database"""
        return await PlayerRow.exists().where((PlayerRow.id == self.id) & (PlayerRow.bot == self.bot))

    @maybe_cached
    async def fetch_volume(self) -> int:
        """Fetch the volume of the player from the db"""

        player = (
            await PlayerRow.select(PlayerRow.volume)
            .where((PlayerRow.id == self.id) & (PlayerRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )
        return player["volume"] if player else PlayerRow.volume.default

    async def update_volume(self, volume: int) -> None:
        """Update the volume of the player in the db"""
        await PlayerRow.insert(PlayerRow(id=self.id, bot=self.bot, volume=volume)).on_conflict(
            action="DO UPDATE", target=(PlayerRow.id, PlayerRow.bot), values=[PlayerRow.volume]
        )
        await self.update_cache((self.fetch_volume, volume), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_max_volume(self) -> int:
        """Fetch the max volume of the player from the db"""
        player = (
            await PlayerRow.select(PlayerRow.max_volume)
            .where((PlayerRow.id == self.id) & (PlayerRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )
        return player["max_volume"] if player else PlayerRow.max_volume.default

    async def update_max_volume(self, max_volume: int) -> None:
        """Update the max volume of the player in the db"""
        await PlayerRow.insert(PlayerRow(id=self.id, bot=self.bot, max_volume=max_volume)).on_conflict(
            action="DO UPDATE",
            target=(PlayerRow.id, PlayerRow.bot),
            values=[PlayerRow.max_volume],
        )
        await self.update_cache((self.fetch_max_volume, max_volume), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_auto_play_playlist_id(self) -> int:
        """Fetch the auto play playlist ID of the player"""
        player = (
            await PlayerRow.select(PlayerRow.auto_play_playlist_id)
            .where((PlayerRow.id == self.id) & (PlayerRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )
        return player["auto_play_playlist_id"] if player else PlayerRow.auto_play_playlist_id.default

    async def update_auto_play_playlist_id(self, auto_play_playlist_id: int) -> None:
        """Update the auto play playlist ID of the player"""
        await PlayerRow.insert(
            PlayerRow(id=self.id, bot=self.bot, auto_play_playlist_id=auto_play_playlist_id)
        ).on_conflict(
            action="DO UPDATE",
            target=(PlayerRow.id, PlayerRow.bot),
            values=[PlayerRow.auto_play_playlist_id],
        )
        await self.update_cache((self.fetch_auto_play_playlist_id, auto_play_playlist_id), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_text_channel_id(self) -> int:
        """Fetch the text channel ID of the player"""
        player = (
            await PlayerRow.select(PlayerRow.text_channel_id)
            .where((PlayerRow.id == self.id) & (PlayerRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )
        return player["text_channel_id"] if player else PlayerRow.text_channel_id.default

    async def update_text_channel_id(self, text_channel_id: int) -> None:
        """Update the text channel ID of the player"""
        await PlayerRow.insert(PlayerRow(id=self.id, bot=self.bot, text_channel_id=text_channel_id)).on_conflict(
            action="DO UPDATE",
            target=(PlayerRow.id, PlayerRow.bot),
            values=[PlayerRow.text_channel_id],
        )
        await self.update_cache((self.fetch_text_channel_id, text_channel_id), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_notify_channel_id(self) -> int:
        """Fetch the notify channel ID of the player"""
        player = (
            await PlayerRow.select(PlayerRow.notify_channel_id)
            .where((PlayerRow.id == self.id) & (PlayerRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )
        return player["notify_channel_id"] if player else PlayerRow.notify_channel_id.default

    async def update_notify_channel_id(self, notify_channel_id: int) -> None:
        """Update the notify channel ID of the player"""
        await PlayerRow.insert(PlayerRow(id=self.id, bot=self.bot, notify_channel_id=notify_channel_id)).on_conflict(
            action="DO UPDATE",
            target=(PlayerRow.id, PlayerRow.bot),
            values=[PlayerRow.notify_channel_id],
        )
        await self.update_cache((self.fetch_notify_channel_id, notify_channel_id), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_forced_channel_id(self) -> int:
        """Fetch the forced channel ID of the player"""
        player = (
            await PlayerRow.select(PlayerRow.forced_channel_id)
            .where((PlayerRow.id == self.id) & (PlayerRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )
        return player["forced_channel_id"] if player else PlayerRow.forced_channel_id.default

    async def update_forced_channel_id(self, forced_channel_id: int) -> None:
        """Update the forced channel ID of the player"""
        await PlayerRow.insert(PlayerRow(id=self.id, bot=self.bot, forced_channel_id=forced_channel_id)).on_conflict(
            action="DO UPDATE",
            target=(PlayerRow.id, PlayerRow.bot),
            values=[PlayerRow.forced_channel_id],
        )
        await self.update_cache((self.fetch_forced_channel_id, forced_channel_id), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_repeat_current(self) -> bool:
        """Fetch the repeat current of the player"""
        player = (
            await PlayerRow.select(PlayerRow.repeat_current)
            .where((PlayerRow.id == self.id) & (PlayerRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )
        return player["repeat_current"] if player else PlayerRow.repeat_current.default

    async def update_repeat_current(self, repeat_current: bool) -> None:
        """Update the repeat current of the player"""
        await PlayerRow.insert(PlayerRow(id=self.id, bot=self.bot, repeat_current=repeat_current)).on_conflict(
            action="DO UPDATE",
            target=(PlayerRow.id, PlayerRow.bot),
            values=[PlayerRow.repeat_current],
        )
        await self.update_cache((self.fetch_repeat_current, repeat_current), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_repeat_queue(self) -> bool:
        """Fetch the repeat queue of the player"""
        player = (
            await PlayerRow.select(PlayerRow.repeat_queue)
            .where((PlayerRow.id == self.id) & (PlayerRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )
        return player["repeat_queue"] if player else PlayerRow.repeat_queue.default

    async def update_repeat_queue(self, repeat_queue: bool) -> None:
        """Update the repeat queue of the player"""
        await PlayerRow.insert(PlayerRow(id=self.id, bot=self.bot, repeat_queue=repeat_queue)).on_conflict(
            action="DO UPDATE",
            target=(PlayerRow.id, PlayerRow.bot),
            values=[PlayerRow.repeat_queue],
        )
        await self.update_cache((self.fetch_repeat_queue, repeat_queue), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_shuffle(self) -> bool:
        """Fetch the shuffle of the player"""
        player = (
            await PlayerRow.select(PlayerRow.shuffle)
            .where((PlayerRow.id == self.id) & (PlayerRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )
        return player["shuffle"] if player else PlayerRow.shuffle.default

    async def update_shuffle(self, shuffle: bool) -> None:
        """Update the shuffle of the player"""
        await PlayerRow.insert(PlayerRow(id=self.id, bot=self.bot, shuffle=shuffle)).on_conflict(
            action="DO UPDATE",
            target=(PlayerRow.id, PlayerRow.bot),
            values=[PlayerRow.shuffle],
        )
        await self.update_cache((self.fetch_shuffle, shuffle), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_auto_shuffle(self) -> bool:
        """Fetch the auto shuffle of the player"""
        player = (
            await PlayerRow.select(PlayerRow.auto_shuffle)
            .where((PlayerRow.id == self.id) & (PlayerRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )
        return player["auto_shuffle"] if player else PlayerRow.auto_shuffle.default

    async def update_auto_shuffle(self, auto_shuffle: bool) -> None:
        """Update the auto shuffle of the player"""
        await PlayerRow.insert(PlayerRow(id=self.id, bot=self.bot, auto_shuffle=auto_shuffle)).on_conflict(
            action="DO UPDATE",
            target=(PlayerRow.id, PlayerRow.bot),
            values=[PlayerRow.auto_shuffle],
        )
        await self.update_cache((self.fetch_auto_shuffle, auto_shuffle), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_auto_play(self) -> bool:
        """Fetch the auto play of the player"""
        player = (
            await PlayerRow.select(PlayerRow.auto_play)
            .where((PlayerRow.id == self.id) & (PlayerRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )
        return player["auto_play"] if player else PlayerRow.auto_play.default

    async def update_auto_play(self, auto_play: bool) -> None:
        """Update the auto play of the player"""
        await PlayerRow.insert(PlayerRow(id=self.id, bot=self.bot, auto_play=auto_play)).on_conflict(
            action="DO UPDATE",
            target=(PlayerRow.id, PlayerRow.bot),
            values=[PlayerRow.auto_play],
        )
        await self.update_cache((self.fetch_auto_play, auto_play), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_self_deaf(self) -> bool:
        """Fetch the self deaf of the player"""
        player = (
            await PlayerRow.select(PlayerRow.self_deaf)
            .where((PlayerRow.id == self.id) & (PlayerRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )
        return player["self_deaf"] if player else PlayerRow.self_deaf.default

    async def update_self_deaf(self, self_deaf: bool) -> None:
        """Update the self deaf of the player"""
        await PlayerRow.insert(PlayerRow(id=self.id, bot=self.bot, self_deaf=self_deaf)).on_conflict(
            action="DO UPDATE",
            target=(PlayerRow.id, PlayerRow.bot),
            values=[PlayerRow.self_deaf],
        )
        await self.update_cache((self.fetch_self_deaf, self_deaf), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_extras(self) -> JSON_DICT_TYPE:
        """Fetch the extras of the player"""
        player = (
            await PlayerRow.select(PlayerRow.extras)
            .where((PlayerRow.id == self.id) & (PlayerRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )
        return player["extras"] if player else json.loads(PlayerRow.extras.default)

    async def update_extras(self, extras: JSON_DICT_TYPE) -> None:
        """Update the extras of the player"""
        await PlayerRow.insert(PlayerRow(id=self.id, bot=self.bot, extras=extras)).on_conflict(
            action="DO UPDATE", target=(PlayerRow.id, PlayerRow.bot), values=[PlayerRow.extras]
        )
        await self.update_cache((self.fetch_extras, extras), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_effects(
        self,
    ) -> dict[str, int | None | dict[str, int | float | list[dict[str, float | None]] | None]]:
        """Fetch the effects of the player"""
        player = (
            await PlayerRow.select(PlayerRow.effects)
            .where((PlayerRow.id == self.id) & (PlayerRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )
        return player["effects"] if player else json.loads(PlayerRow.effects.default)

    async def update_effects(
        self, effects: dict[str, int | None | dict[str, int | float | list[dict[str, float | None]] | None]]
    ) -> None:
        """Update the effects of the player"""
        await PlayerRow.insert(PlayerRow(id=self.id, bot=self.bot, effects=effects)).on_conflict(
            action="DO UPDATE",
            target=(PlayerRow.id, PlayerRow.bot),
            values=[PlayerRow.effects],
        )
        await self.update_cache((self.fetch_effects, effects), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_empty_queue_dc(self) -> TimedFeature:
        """Fetch the empty queue dc of the player"""
        player = (
            await PlayerRow.select(PlayerRow.empty_queue_dc)
            .where((PlayerRow.id == self.id) & (PlayerRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )
        return TimedFeature.from_dict(
            player["empty_queue_dc"] if player else json.loads(PlayerRow.empty_queue_dc.default)
        )

    async def update_empty_queue_dc(self, empty_queue_dc: dict[str, int | bool]) -> None:
        """Update the empty queue dc of the player"""
        await PlayerRow.insert(PlayerRow(id=self.id, bot=self.bot, empty_queue_dc=empty_queue_dc)).on_conflict(
            action="DO UPDATE",
            target=(PlayerRow.id, PlayerRow.bot),
            values=[PlayerRow.empty_queue_dc],
        )
        await self.update_cache(
            (self.fetch_empty_queue_dc, TimedFeature.from_dict(empty_queue_dc)), (self.exists, True)
        )
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_alone_dc(self) -> TimedFeature:
        """Fetch the alone dc of the player"""
        player = (
            await PlayerRow.select(PlayerRow.alone_dc)
            .where((PlayerRow.id == self.id) & (PlayerRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )
        return TimedFeature.from_dict(player["alone_dc"] if player else json.loads(PlayerRow.alone_dc.default))

    async def update_alone_dc(self, alone_dc: dict[str, int | bool]) -> None:
        """Update the alone dc of the player"""
        await PlayerRow.insert(PlayerRow(id=self.id, bot=self.bot, alone_dc=alone_dc)).on_conflict(
            action="DO UPDATE",
            target=(PlayerRow.id, PlayerRow.bot),
            values=[PlayerRow.alone_dc],
        )
        await self.update_cache((self.fetch_alone_dc, TimedFeature.from_dict(alone_dc)), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_alone_pause(self) -> TimedFeature:
        """Fetch the alone pause of the player"""
        player = (
            await PlayerRow.select(PlayerRow.alone_pause)
            .where((PlayerRow.id == self.id) & (PlayerRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )
        return TimedFeature.from_dict(player["alone_pause"] if player else json.loads(PlayerRow.alone_pause.default))

    async def update_alone_pause(self, alone_pause: dict[str, int | bool]) -> None:
        """Update the alone pause of the player"""
        await PlayerRow.insert(PlayerRow(id=self.id, bot=self.bot, alone_pause=alone_pause)).on_conflict(
            action="DO UPDATE",
            target=(PlayerRow.id, PlayerRow.bot),
            values=[PlayerRow.alone_pause],
        )
        await self.update_cache((self.fetch_alone_pause, TimedFeature.from_dict(alone_pause)), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_dj_users(self) -> set[int]:
        """Fetch the disc jockey users of the player"""
        player = (
            await PlayerRow.select(PlayerRow.dj_users)
            .where((PlayerRow.id == self.id) & (PlayerRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )

        return set(player["dj_users"] if player else PlayerRow.dj_users.default)

    async def add_to_dj_users(self, user: discord.Member) -> None:
        """Add a user to the disc jockey users of the player"""
        await PlayerRow.insert(PlayerRow(id=self.id, bot=self.bot, dj_users=[user.id])).on_conflict(
            action="DO UPDATE",
            target=(PlayerRow.id, PlayerRow.bot),
            values=[PlayerRow.dj_users, QueryString("array_cat(player.dj_users, EXCLUDED.dj_users)")],
        )
        await self.update_cache((self.exists, True))
        await self.invalidate_cache(self.fetch_all, self.fetch_dj_users)

    async def remove_from_dj_users(self, user: discord.Member) -> None:
        """Remove a user from the disc jockey users of the player"""
        await PlayerRow.update(dj_users=QueryString("array_remove(dj_users, {})", user.id)).where(
            PlayerRow.id == self.id & PlayerRow.bot == self.bot
        )
        await self.update_cache((self.exists, True))
        await self.invalidate_cache(self.fetch_all, self.fetch_dj_users)

    async def bulk_add_dj_users(self, *users: discord.Member) -> None:
        """Add disc jockey users to the player"""
        if not users:
            return
        await PlayerRow.insert(PlayerRow(id=self.id, bot=self.bot, dj_users=[u.id for u in users])).on_conflict(
            action="DO UPDATE",
            target=(PlayerRow.id, PlayerRow.bot),
            values=[PlayerRow.dj_users, QueryString("array_cat(player.dj_users, EXCLUDED.dj_users)")],
        )
        await self.update_cache((self.exists, True))
        await self.invalidate_cache(self.fetch_all, self.fetch_dj_users)

    async def bulk_remove_dj_users(self, *users: discord.Member) -> None:
        """Remove disc jockey users from the player.

        Parameters
        ----------
        users : discord.Member
            The users to add
        """
        if not users:
            return
        for user in users:
            await self.remove_from_dj_users(user)

    async def dj_users_reset(self) -> None:
        """Reset the disc jockey users of the player"""
        await PlayerRow.insert(PlayerRow(id=self.id, bot=self.bot, dj_users=[])).on_conflict(
            action="DO UPDATE",
            target=(PlayerRow.id, PlayerRow.bot),
            values=[PlayerRow.dj_users],
        )
        await self.update_cache((self.fetch_dj_users, set()), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_dj_roles(self) -> set[int]:
        """Fetch the disc jockey roles of the player"""
        player = (
            await PlayerRow.select(PlayerRow.dj_roles)
            .where((PlayerRow.id == self.id) & (PlayerRow.bot == self.bot))
            .first()
            .output(load_json=True, nested=True)
        )

        return set(player["dj_roles"] if player else PlayerRow.dj_roles.default)

    async def add_to_dj_roles(self, role: discord.Role) -> None:
        """Add disc jockey roles to the player"""
        await PlayerRow.insert(PlayerRow(id=self.id, bot=self.bot, dj_roles=[role.id])).on_conflict(
            action="DO UPDATE",
            target=(PlayerRow.id, PlayerRow.bot),
            values=[PlayerRow.dj_roles, QueryString("array_cat(player.dj_roles, EXCLUDED.dj_roles)")],
        )
        await self.update_cache((self.exists, True))
        await self.invalidate_cache(self.fetch_all, self.fetch_dj_roles)

    async def remove_from_dj_roles(self, role: discord.Role) -> None:
        """Remove disc jockey roles from the player"""
        await PlayerRow.update(dj_roles=QueryString("array_remove(dj_roles, {})", role.id)).where(
            PlayerRow.id == self.id & PlayerRow.bot == self.bot
        )
        await self.update_cache((self.exists, True))
        await self.invalidate_cache(self.fetch_all, self.fetch_dj_roles)

    async def bulk_add_dj_roles(self, *roles: discord.Role) -> None:
        """Add disc jockey roles to the player.

        Parameters
        ----------
        roles : discord.Role
            The roles to add
        """
        if not roles:
            return
        await PlayerRow.insert(PlayerRow(id=self.id, bot=self.bot, dj_roles=[r.id for r in roles])).on_conflict(
            action="DO UPDATE",
            target=(PlayerRow.id, PlayerRow.bot),
            values=[PlayerRow.dj_roles, QueryString("array_cat(player.dj_roles, EXCLUDED.dj_roles)")],
        )
        await self.update_cache((self.exists, True))
        await self.invalidate_cache(self.fetch_all, self.fetch_dj_roles)

    async def bulk_remove_dj_roles(self, *roles: discord.Role) -> None:
        """Remove disc jockey roles from the player.

        Parameters
        ----------
        roles : discord.Role
            The roles to add.
        """
        if not roles:
            return
        for role in roles:
            await self.remove_from_dj_roles(role)

    async def dj_roles_reset(self) -> None:
        """Reset the disc jockey roles of the player"""
        await PlayerRow.insert(PlayerRow(id=self.id, bot=self.bot, dj_roles=[])).on_conflict(
            action="DO UPDATE",
            target=(PlayerRow.id, PlayerRow.bot),
            values=[PlayerRow.dj_roles],
        )
        await self.update_cache((self.fetch_dj_roles, set()), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    async def _role_id_in_dj_roles(self, role_id: int) -> bool:
        return await PlayerRow.exists().where(
            (PlayerRow.id == self.id) & (PlayerRow.bot == self.bot) & PlayerRow.dj_roles.any(role_id)
        )

    async def _userid_in_dj_users(self, user_id: int) -> bool:
        return await PlayerRow.exists().where(
            (PlayerRow.id == self.id) & (PlayerRow.bot == self.bot) & PlayerRow.dj_users.any(user_id)
        )

    async def is_dj(
        self,
        user: discord.Member,
        *,
        additional_role_ids: list | None = None,
        additional_user_ids: list | None = None,
    ) -> bool:
        """Check if a user is a disc jockey.

        Parameters
        ----------
        user : discord.Member
            The user to check.
        additional_role_ids : list
            The additional disc jockey role ids to check.
        additional_user_ids : list
            The additional disc jockey user ids to check.

        Returns
        -------
        bool
            Whether the user is a disc jockey.
        """
        if additional_user_ids and user.id in additional_user_ids:
            return True
        if additional_role_ids and any(r.id in additional_role_ids for r in user.roles):
            return True
        if __ := user.guild:
            bot = self.client.bot
            if hasattr(bot, "is_owner") and await bot.is_owner(typing.cast(discord.User, user)):
                return True
            if hasattr(bot, "is_admin") and await bot.is_admin(user):
                return True
            if hasattr(bot, "is_mod") and await bot.is_mod(user):
                return True
        if await self._userid_in_dj_users(user.id):
            return True
        dj_roles = await self.fetch_dj_roles()
        return any(r.id in dj_roles for r in user.roles) if dj_roles else True



from __future__ import annotations

import asyncio

import discord

from pylav.extension.red.ui.selectors.options.generic import EntryOption
from pylav.helpers.format.strings import shorten_string
from pylav.type_hints.bot import DISCORD_COG_TYPE, DISCORD_INTERACTION_TYPE
from pylav.type_hints.generics import ANY_GENERIC_TYPE


class EntrySelectSelector(discord.ui.Select):
    def __init__(
        self,
        options: list[EntryOption],
        cog: DISCORD_COG_TYPE,
        placeholder: str,
        mapping: dict[str, ANY_GENERIC_TYPE],
    ):
        super().__init__(
            min_values=1,
            max_values=1,
            options=options,
            placeholder=shorten_string(max_length=100, string=placeholder),
        )
        self.cog = cog
        self.mapping = mapping
        self.entry: ANY_GENERIC_TYPE = None
        self.responded = asyncio.Event()

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE) -> None:
        entry_id = self.values[0]
        self.entry: ANY_GENERIC_TYPE = self.mapping.get(entry_id)
        self.responded.set()
        self.view.stop()
        await self.view.on_timeout()



from __future__ import annotations

import asyncio
import ssl
import urllib.request

from pylav.extension.m3u.parser import urljoin


def parsed_url(url: str) -> str | bytes:
    return urljoin(url, ".")


class DefaultHTTPClient:
    __slots__ = ("proxies",)

    def __init__(self, proxies: dict[str, list[str]] = None) -> None:
        self.proxies = proxies

    async def download(
        self, uri: str, timeout: float | None = None, headers: dict[str, str] | None = None, verify_ssl: bool = True
    ) -> tuple[str, str]:
        if headers is None:
            headers = {}
        proxy_handler = await asyncio.to_thread(urllib.request.ProxyHandler, self.proxies)
        https_handler = await asyncio.to_thread(HTTPSHandler, verify_ssl=verify_ssl)
        opener = await asyncio.to_thread(urllib.request.build_opener, proxy_handler, https_handler)
        opener.addheaders = headers.items()
        resource = await asyncio.to_thread(opener.open, uri, timeout=timeout)
        base_uri = parsed_url(await asyncio.to_thread(resource.geturl))
        content = (await asyncio.to_thread(resource.read)).decode(
            await asyncio.to_thread(resource.headers.get_content_charset, failobj="utf-8")
        )
        return content, base_uri


class HTTPSHandler:
    __slots__ = ()

    def __new__(cls, verify_ssl: bool = True) -> urllib.request.HTTPSHandler:
        context = ssl.create_default_context()
        if not verify_ssl:
            context.check_hostname = False
            context.verify_mode = ssl.CERT_NONE
        return urllib.request.HTTPSHandler(context=context)



from __future__ import annotations

import dataclasses
import time
from collections.abc import Iterator
from typing import Any

from discord.backoff import ExponentialBackoff

from pylav.type_hints.generics import ANY_GENERIC_TYPE


class MissingSentinel(str):
    """A sentinel class for missing values."""

    def __str__(self) -> str:
        return "MISSING"

    def __repr__(self) -> str:
        return "MISSING"

    def __bool__(self) -> bool:
        return False

    def __int__(self) -> int:
        return 0

    def __float__(self) -> float:
        return 0.0

    def __len__(self) -> int:
        return 0

    def __iter__(self) -> Iterator:
        return iter([])

    def __getitem__(self, item) -> None:
        return None

    def __getattr__(self, item) -> None:
        return None

    def __divmod__(self, other) -> tuple[int, int]:
        return 0, 0

    def __rdivmod__(self, other) -> tuple[int, int]:
        return 0, 0

    def __floor__(self) -> int:
        return 0

    def __ceil__(self) -> int:
        return 0

    def __round__(self) -> int:
        return 0

    def __trunc__(self) -> int:
        return 0

    def __add__(self, other: ANY_GENERIC_TYPE) -> ANY_GENERIC_TYPE:
        return other

    def __radd__(self, other: ANY_GENERIC_TYPE) -> ANY_GENERIC_TYPE:
        return other

    def __sub__(self, other: ANY_GENERIC_TYPE) -> ANY_GENERIC_TYPE:
        return other

    def __rsub__(self, other: ANY_GENERIC_TYPE) -> ANY_GENERIC_TYPE:
        return other

    def __mul__(self, other: Any) -> int:
        return 0

    def __rmul__(self, other: Any) -> int:
        return 0

    def __matmul__(self, other: Any) -> int:
        return 0

    def __rmatmul__(self, other: Any) -> int:
        return 0

    def __mod__(self, other: Any) -> int:
        return 0

    def __rmod__(self, other: Any) -> int:
        return 0

    def __rdiv__(self, other: Any) -> int:
        return 0

    def __truediv__(self, other: Any) -> int:
        return 0

    def __rtruediv__(self, other: Any) -> int:
        return 0

    def __floordiv__(self, other: Any) -> int:
        return 0

    def __rfloordiv__(self, other: Any) -> int:
        return 0

    def __pow__(self, other: Any) -> int:
        return 0

    def __rpow__(self, other: Any) -> int:
        return 0

    def __lshift__(self, other: Any) -> int:
        return 0

    def __rlshift__(self, other: Any) -> int:
        return 0

    def __le__(self, other: Any) -> bool:
        return True

    def __lt__(self, other: Any) -> bool:
        return True

    def __ge__(self, other: Any) -> bool:
        return False

    def __gt__(self, other: Any) -> bool:
        return False

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, MissingSentinel)

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)

    def __hash__(self) -> int:
        return 0

    def __contains__(self, item: Any) -> bool:
        return False


MISSING: Any = MissingSentinel("MISSING")


@dataclasses.dataclass(eq=True, slots=True, unsafe_hash=True, order=True, kw_only=True, frozen=True)
class TimedFeature:
    """A timed feature data wrapper."""

    enabled: bool = False
    time: int = 60

    def to_dict(self) -> dict[str, bool | int]:
        """Convert to dict"""
        return {"enabled": self.enabled, "time": self.time}

    @classmethod
    def from_dict(cls, data: dict[str, bool | int]) -> TimedFeature:
        """Convert from dict"""
        return cls(enabled=data["enabled"], time=data["time"])


class ExponentialBackoffWithReset(ExponentialBackoff):
    """
    Exponential backoff with reset
    """

    def __init__(self, base: int = 1, *, integral: ANY_GENERIC_TYPE = False) -> None:
        super().__init__(base=base, integral=integral)

    def reset(self) -> None:
        """
        Reset the backoff to its initial state.
        """
        self._last_invocation: float = time.monotonic()
        self._exp = 0



from __future__ import annotations

from dataclasses import dataclass, field
from typing import NotRequired

from pylav.constants.builtin_nodes import BUNDLED_NODES_IDS_HOST_MAPPING
from pylav.constants.node_features import SUPPORTED_FEATURES, SUPPORTED_SOURCES
from pylav.storage.database.cache.model import CachedModel
from pylav.storage.database.tables.nodes import Sessions
from pylav.type_hints.dict_typing import JSON_DICT_TYPE


@dataclass(eq=True, slots=True, unsafe_hash=True, order=True, kw_only=True)
class NodeMock(CachedModel):
    id: int
    data: JSON_DICT_TYPE

    session_id: NotRequired[str | None] = field(repr=False, init=False, hash=None, compare=False, default=...)

    def __post_init__(self):
        self.session_id = ...

    def get_cache_key(self) -> str:
        return f"{self.id}"

    @staticmethod
    async def exists() -> bool:
        """Check if the node exists in the database.

        Returns
        -------
        bool
            Whether the node exists in the database.
        """
        return False

    async def delete(self) -> None:
        """Delete the node from the database"""

    async def fetch_all(self) -> JSON_DICT_TYPE:
        return {"id": self.id, **self.data}

    async def fetch_name(self) -> str:
        """Fetch the node from the database.

        Returns
        -------
        ste
            The node's name.
        """
        return self.data["name"]

    async def update_name(self, name: str) -> None:
        """Update the node's name in the database"""
        self.data["name"] = name

    async def fetch_ssl(self) -> bool:
        """Fetch the node's ssl setting from the database.

        Returns
        -------
        bool
            The node's ssl setting.
        """
        return self.data["ssl"]

    async def update_ssl(self, ssl: bool) -> None:
        """Update the node's ssl setting in the database"""
        self.data["ssl"] = ssl

    async def fetch_resume_timeout(self) -> int:
        """Fetch the node's resume timeout from the database.

        Returns
        -------
        int
            The node's resume timeout.
        """
        return self.data["resume_timeout"]

    async def update_resume_timeout(self, resume_timeout: int) -> None:
        """Update the node's resume timeout in the database"""
        self.data["resume_timeout"] = resume_timeout

    async def fetch_reconnect_attempts(self) -> int:
        """Fetch the node's reconnect attempts from the database.

        Returns
        -------
        int
            The node's reconnect attempts.
        """
        return self.data["reconnect_attempts"]

    async def update_reconnect_attempts(self, reconnect_attempts: int) -> None:
        """Update the node's reconnect attempts in the database"""
        self.data["reconnect_attempts"] = reconnect_attempts

    async def fetch_search_only(self) -> bool:
        """Fetch the node's search only setting from the database.

        Returns
        -------
        bool
            The node's search only setting.
        """
        return self.data["search_only"]

    async def update_search_only(self, search_only: bool) -> None:
        """Update the node's search only setting in the database"""
        self.data["search_only"] = search_only

    async def fetch_managed(self) -> bool:
        """Fetch the node's managed setting from the database.

        Returns
        -------
        bool
            The node's managed setting.
        """
        return self.data["managed"]

    async def update_managed(self, managed: bool) -> None:
        """Update the node's managed setting in the database"""
        self.data["managed"] = managed

    async def fetch_extras(self) -> JSON_DICT_TYPE:
        """Fetch the node's extras from the database.

        Returns
        -------
        dict
            The node's extras.
        """
        return self.data["extras"]

    async def update_extras(self, extras: JSON_DICT_TYPE) -> None:
        """Update the node's extras in the database"""
        self.data["extras"] = extras

    async def fetch_yaml(self) -> JSON_DICT_TYPE:
        """Fetch the node's yaml from the database.

        Returns
        -------
        dict
            The node's yaml.
        """
        return self.data["yaml"]

    async def update_yaml(self, yaml_data: JSON_DICT_TYPE) -> None:
        """Update the node's yaml in the database"""
        self.data["yaml"] = yaml_data

    async def fetch_disabled_sources(self) -> list[str]:
        """Fetch the node's disabled sources from the database.

        Returns
        -------
        list[str]
            The node's disabled sources.
        """
        return self.data["disabled_sources"]

    async def update_disabled_sources(self, disabled_sources: list[str]) -> None:
        """Update the node's disabled sources in the database"""
        source = set(map(str.strip, map(str.lower, disabled_sources)))
        intersection = list(source & SUPPORTED_SOURCES.union(SUPPORTED_FEATURES))
        self.data["disabled_sources"] = intersection

    async def add_to_disabled_sources(self, source: str) -> None:
        """Add a source to the node's disabled sources in the database"""
        data = set(self.data["disabled_sources"])
        data.update([source])
        self.data["disabled_sources"] = list(data)

    async def remove_from_disabled_sources(self, source: str) -> None:
        """Remove a source from the node's disabled sources in the database"""
        data = set(self.data["disabled_sources"])
        data.discard(source)
        self.data["disabled_sources"] = list(data)

    async def bulk_add_to_disabled_sources(self, sources: list[str]) -> None:
        """Add sources to the node's disabled sources in the database"""
        source = set(map(str.strip, map(str.lower, [sources])))
        source.update(self.data["disabled_sources"])
        intersection = list(source & SUPPORTED_SOURCES.union(SUPPORTED_FEATURES))
        self.data["disabled_sources"] = intersection

    async def bulk_remove_from_disabled_sources(self, sources: list[str]) -> None:
        """Remove sources from the node's disabled sources in the database"""
        if not sources:
            return
        for source in sources:
            await self.remove_from_disabled_sources(source)

    async def get_connection_args(self) -> dict[str, int | str | bool | None, list[str]]:
        """Get the connection args for the node.

        Returns
        -------
        dict
            The connection args.
        """
        data = await self.fetch_all()
        if self.id in BUNDLED_NODES_IDS_HOST_MAPPING:
            data["yaml"]["lavalink"]["server"]["password"] = self.client._user_id
        return {
            "unique_identifier": self.id,
            "host": data["yaml"]["server"]["address"],
            "port": data["yaml"]["server"]["port"],
            "password": data["yaml"]["lavalink"]["server"]["password"],
            "name": data["name"],
            "ssl": data["ssl"],
            "reconnect_attempts": data["reconnect_attempts"],
            "search_only": data["search_only"],
            "resume_timeout": data["resume_timeout"],
            "disabled_sources": data["disabled_sources"],
            "managed": data["managed"],
            "temporary": True,
        }

    @classmethod
    async def create_managed(cls, identifier: int) -> None:
        """Create the player in the database"""

    async def fetch_session(self) -> str | None:
        """Fetch the node's session from the database.

        Returns
        -------
        str | None
            The node's session.
        """

        if self.session_id is ...:
            session = (
                await Sessions.select(Sessions.id)
                .where((Sessions.node == self.id) & (Sessions.bot == self.client.bot.user.id))  # noqa: E712
                .first()
                .output(load_json=True)
            )
            if session and (sid := session.get("id")):
                self.session_id = sid
                return self.session_id
            self.session_id = None
        return self.data.get("session", None)

    async def update_session(self, session: str | None) -> None:
        """Update the node's session in the database"""
        self.data["session"] = session
        self.session_id = session
        await Sessions.insert(Sessions(id=session, node=self.id, bot=self.client.bot.user.id)).on_conflict(
            action="DO UPDATE",
            target=(Sessions.node, Sessions.bot),
            values=[Sessions.id],
        )



from __future__ import annotations

import datetime
from typing import TYPE_CHECKING

import aiohttp
import aiohttp_client_cache

from pylav.compat import json
from pylav.constants.config import REDIS_FULL_ADDRESS_RESPONSE_CACHE
from pylav.extension.flowery.lyrics import LyricsAPI
from pylav.utils.aiohttp_postgres_cache import PostgresCacheBackend

if TYPE_CHECKING:
    from pylav.core.client import Client


class FloweryAPI:
    """A wrapper for the Flowery API."""

    def __init__(self, client: Client) -> None:
        if REDIS_FULL_ADDRESS_RESPONSE_CACHE:
            self._aiohttp_client_cache = aiohttp_client_cache.RedisBackend(
                address=REDIS_FULL_ADDRESS_RESPONSE_CACHE,
                cache_control=True,
                allowed_codes=(200,),
                allowed_methods=("GET",),
                include_headers=True,
                expire_after=datetime.timedelta(days=1),
                timeout=2.5,
            )
        else:
            self._aiohttp_client_cache = PostgresCacheBackend(
                cache_control=True,
                allowed_codes=(200,),
                allowed_methods=("GET",),
                include_headers=True,
                expire_after=datetime.timedelta(days=1),
                timeout=2.5,
            )
        self._cached_session = aiohttp_client_cache.CachedSession(
            timeout=aiohttp.ClientTimeout(total=30),
            json_serialize=json.dumps,
            cache=self._aiohttp_client_cache,
            headers={"User-Agent": f"PyLav/{client.lib_version} (https://github.com/PyLav/PyLav)"},
        )

        self._lyrics = LyricsAPI(client, self)

    async def shutdown(self) -> None:
        """Close the cached session."""
        await self._cached_session.close()

    @property
    def lyrics(self) -> LyricsAPI:
        """The lyrics API client."""
        return self._lyrics



from __future__ import annotations

from typing import TYPE_CHECKING

import discord

from pylav.events.base import PyLavEvent
from pylav.events.track.track_start import TrackStartEvent as TrackStartEvent
from pylav.nodes.api.responses.websocket import TrackEnd, TrackException, TrackStuck

if TYPE_CHECKING:
    from pylav.nodes.node import Node
    from pylav.players.player import Player
    from pylav.players.tracks.obj import Track


class TrackStuckEvent(PyLavEvent):
    """This event is dispatched when the currently playing track is stuck.
    This normally has something to do with the stream you are playing
    and not Lavalink itself.

    Event can be listened to by adding a listener with the name `pylav_track_stuck_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that is stuck.
    track: :class:`Track`
        The track that is stuck.
    threshold: :class:`int`
        The threshold in milliseconds.
    node: :class:`Node`
        The node that dispatched the event.
    event: :class:`TrackStuck`
        The raw event object.

    """

    __slots__ = ("player", "track", "threshold", "node", "event")

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackStuck):
        self.player = player
        self.track = track
        self.threshold = event_object.thresholdMs
        self.node = node
        self.event = event_object


class TrackExceptionEvent(PyLavEvent):
    """This event is dispatched when an exception occurs while playing a track.

    Event can be listened to by adding a listener with the name `pylav_track_exception_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that encountered the exception.
    track: :class:`Track`
        The track that encountered the exception.
    exception: :class:`str`
        The exception that was encountered.
    node: :class:`Node`
        The node that dispatched the event.
    event: :class:`TrackException`
        The raw event object.
    """

    __slots__ = ("player", "track", "exception", "node", "event")

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackException):
        self.player = player
        self.track = track
        self.exception = event_object.exception.cause
        self.node = node
        self.event = event_object


class TrackEndEvent(PyLavEvent):
    """This event is dispatched when the player finished playing a track.

    Event can be listened to by adding a listener with the name `pylav_track_end_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that finished playing the track.
    track: :class:`Track`
        The track that finished playing.
    reason: :class:`str`
        The reason why the track finished playing.
    node: :class:`Node`
        The node that dispatched the event.
    event: :class:`TrackEnd`
        The raw event object.

    """

    __slots__ = ("player", "track", "reason", "node", "event")

    def __init__(self, player: Player, track: Track, node: Node, event_object: TrackEnd) -> None:
        self.player = player
        self.track = track
        self.reason = event_object.reason
        self.node = node
        self.event = event_object


class TrackAutoPlayEvent(PyLavEvent):
    """This event is dispatched when the player starts to play a track.

    Event can be listened to by adding a listener with the name `pylav_track_start_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that started playing the track.
    track: :class:`Track`
        The track that started playing.

    """

    __slots__ = ("player", "track")

    def __init__(self, player: Player, track: Track) -> None:
        self.player = player
        self.track = track


class TrackResumedEvent(PyLavEvent):
    """This event is dispatched when the player resumes playing a track.

    Event can be listened to by adding a listener with the name `pylav_track_resume_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that resumed playing the track.
    track: :class:`Track`
        The track that resumed playing.
    requester: :class:`discord.Member`
        The member that requested the resume.

    """

    __slots__ = ("player", "track", "requester")

    def __init__(self, player: Player, track: Track, requester: discord.Member) -> None:
        self.player = player
        self.track = track
        self.requester = requester


class TrackSeekEvent(PyLavEvent):
    """This event is dispatched when the player is moved.

    Event can be listened to by adding a listener with the name `pylav_track_seek_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that moved.
    track: :class:`Track`
        The track that was moved.
    requester: :class:`discord.Member`
        The member that requested the move.
    before: :class:`float`
        The position before the move.
    after: :class:`float`
        The position after the move.

    """

    __slots__ = ("player", "track", "requester", "before", "after")

    def __init__(self, player: Player, requester: discord.Member, track: Track, before: float, after: float) -> None:
        self.player = player
        self.requester = requester
        self.track = track
        self.before = before
        self.after = after


class TrackPreviousRequestedEvent(PyLavEvent):
    """This event is dispatched when a history track is requested.

    Event can be listened to by adding a listener with the name `pylav_track_previous_requested_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that requested the track.
    track: :class:`Track`
        The track that was requested.
    requester: :class:`discord.Member`
        The member that requested the track.

    """

    __slots__ = ("player", "track", "requester")

    def __init__(self, player: Player, requester: discord.Member, track: Track) -> None:
        self.player = player
        self.requester = requester
        self.track = track


class TrackSkippedEvent(PyLavEvent):
    """This event is dispatched when a track is skipped.

    Event can be listened to by adding a listener with the name `pylav_track_skipped_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that skipped the track.
    track: :class:`Track`
        The track that was skipped.
    requester: :class:`discord.Member`
        The member that requested the skip.

    """

    __slots__ = ("player", "track", "requester", "position")

    def __init__(self, player: Player, requester: discord.Member, track: Track, position: float) -> None:
        self.player = player
        self.track = track
        self.requester = requester
        self.position = position



from __future__ import annotations

from typing import TYPE_CHECKING

from pylav.events.base import PyLavEvent

if TYPE_CHECKING:
    from pylav.nodes.api.responses.rest_api import LoadTrackResponses
    from pylav.nodes.node import Node


class LavalinkLoadtracksEvent(PyLavEvent):
    """This event is dispatched when the Lavalink /loadtracks endpoint is called.

    Event can be listened to by adding a listener with the name `pylav_lavalink_loadtracks_event`.

    Attributes
    ----------
    node: :class:`Node`
        The node that called the endpoint.

    Parameters
    ----------
    player: :class:`Player`
        The player that has no more songs in queue.
    response: :class:`LoadTrackResponses`
        The response from the endpoint.
    """

    __slots__ = ("player", "response")

    def __init__(self, node: Node, response: LoadTrackResponses) -> None:
        self.node = node
        self.response = response



from __future__ import annotations

from packaging.version import Version

from pylav.constants.node import MAX_SUPPORTED_API_MAJOR_VERSION

VERSION_0_0_0 = Version("0.0.0")
VERSION_3_6_0 = Version("3.6.0")
VERSION_3_7_0 = Version("3.7.0")
VERSION_4_0_0 = Version("4.0.0")
VERSION_5_0_0 = Version("5.0.0")

VERSION_3_7_0_FIRST = Version("3.7.0-alpha.dev")
VERSION_4_0_0_FIRST = Version("4.0.0-alpha.dev")
VERSION_5_0_0_FIRST = Version("5.0.0-alpha.dev")

API_DEVELOPMENT_VERSION = Version(f"{MAX_SUPPORTED_API_MAJOR_VERSION}.999.0-alpha")

# Migration versions
VERSION_0_0_0_2 = Version("0.0.0.post2.dev0")
VERSION_0_3_2_0 = Version("0.3.2.dev")
VERSION_0_3_3_0 = Version("0.3.3.dev")
VERSION_0_3_4_0 = Version("0.3.4.dev")
VERSION_0_3_5_0 = Version("0.3.5.dev")
VERSION_0_3_6_0 = Version("0.3.6.dev")
VERSION_0_7_6_0 = Version("0.7.6.dev")
VERSION_0_8_5_0 = Version("0.8.5.dev")
VERSION_0_8_8_0 = Version("0.8.8.dev")
VERSION_0_9_2_0 = Version("0.9.2.dev")
VERSION_0_10_5_0 = Version("0.10.5.dev")
VERSION_0_11_3_0 = Version("0.11.3.dev")
VERSION_0_11_8_0 = Version("0.11.8.dev")
VERSION_1_0_0 = Version("1.0.0.dev")
VERSION_1_1_17 = Version("1.1.17.dev")
VERSION_1_10_0 = Version("1.10.0.dev")
VERSION_1_10_1 = Version("1.10.1.dev")



from __future__ import annotations

from dataclasses import dataclass

from pylav.compat import json
from pylav.storage.database.cache.model import CachedModel
from pylav.storage.database.tables.player_state import PlayerStateRow
from pylav.type_hints.dict_typing import JSON_DICT_TYPE


@dataclass(eq=True)
class PlayerState(CachedModel):
    """A class to represent the state of a player in the database"""

    id: int
    bot: int
    channel_id: int
    volume: int
    position: float
    auto_play_playlist_id: int | None
    text_channel_id: int | None
    notify_channel_id: int | None
    forced_channel_id: int | None

    paused: bool
    repeat_current: bool
    repeat_queue: bool
    shuffle: bool
    auto_shuffle: bool
    auto_play: bool
    playing: bool
    effect_enabled: bool
    self_deaf: bool

    current: JSON_DICT_TYPE | None
    queue: list[JSON_DICT_TYPE]
    history: list[JSON_DICT_TYPE]
    effects: JSON_DICT_TYPE
    extras: JSON_DICT_TYPE
    pk: None = None

    def get_cache_key(self) -> str:
        """Get the cache key for the player state."""
        return f"{self.id}:{self.bot}:{self.channel_id}"

    def __post_init__(self) -> None:
        if isinstance(self.current, str):
            self.current = json.loads(self.current)
        if isinstance(self.queue, str):
            self.queue = json.loads(self.queue)
        if isinstance(self.history, str):
            self.history = json.loads(self.history)
        if isinstance(self.effects, str):
            self.effects = json.loads(self.effects)
        if isinstance(self.extras, str):
            self.extras = json.loads(self.extras)

    async def delete(self) -> None:
        """Delete the player state from the database"""
        await PlayerStateRow.delete().where((PlayerStateRow.id == self.id) & (PlayerStateRow.bot == self.bot))

    async def save(self) -> None:
        """Save the player state to the database"""
        await PlayerStateRow.insert(
            PlayerStateRow(
                id=self.id,
                bot=self.bot,
                channel_id=self.channel_id,
                volume=self.volume,
                position=self.position,
                auto_play_playlist_id=self.auto_play_playlist_id,
                forced_channel_id=self.forced_channel_id,
                text_channel_id=self.text_channel_id,
                notify_channel_id=self.notify_channel_id,
                paused=self.paused,
                repeat_current=self.repeat_current,
                repeat_queue=self.repeat_queue,
                shuffle=self.shuffle,
                auto_shuffle=self.auto_shuffle,
                auto_play=self.auto_play,
                playing=self.playing,
                effect_enabled=self.effect_enabled,
                self_deaf=self.self_deaf,
                current=self.current,
                queue=self.queue,
                history=self.history,
                effects=self.effects,
                extras=self.extras,
            )
        ).on_conflict(
            action="DO UPDATE",
            target=(PlayerStateRow.id, PlayerStateRow.bot),
            values=[
                PlayerStateRow.channel_id,
                PlayerStateRow.volume,
                PlayerStateRow.position,
                PlayerStateRow.auto_play_playlist_id,
                PlayerStateRow.forced_channel_id,
                PlayerStateRow.text_channel_id,
                PlayerStateRow.notify_channel_id,
                PlayerStateRow.paused,
                PlayerStateRow.repeat_current,
                PlayerStateRow.repeat_queue,
                PlayerStateRow.shuffle,
                PlayerStateRow.auto_shuffle,
                PlayerStateRow.auto_play,
                PlayerStateRow.playing,
                PlayerStateRow.effect_enabled,
                PlayerStateRow.self_deaf,
                PlayerStateRow.current,
                PlayerStateRow.queue,
                PlayerStateRow.history,
                PlayerStateRow.effects,
                PlayerStateRow.extras,
            ],
        )

    @classmethod
    async def get(cls, bot_id: int, guild_id: int) -> PlayerState | None:
        """Get the player state from the database.

        Parameters
        ----------
        bot_id : int
            The bot ID.
        guild_id : int
            The guild ID.

        Returns
        -------
        PlayerState | None
            The player state if found, otherwise None.
        """
        player = (
            await PlayerStateRow.select(
                PlayerStateRow.id,
                PlayerStateRow.bot,
                PlayerStateRow.channel_id,
                PlayerStateRow.volume,
                PlayerStateRow.position,
                PlayerStateRow.auto_play_playlist_id,
                PlayerStateRow.forced_channel_id,
                PlayerStateRow.text_channel_id,
                PlayerStateRow.notify_channel_id,
                PlayerStateRow.paused,
                PlayerStateRow.repeat_current,
                PlayerStateRow.repeat_queue,
                PlayerStateRow.shuffle,
                PlayerStateRow.auto_shuffle,
                PlayerStateRow.auto_play,
                PlayerStateRow.playing,
                PlayerStateRow.effect_enabled,
                PlayerStateRow.self_deaf,
                PlayerStateRow.current,
                PlayerStateRow.queue,
                PlayerStateRow.history,
                PlayerStateRow.effects,
                PlayerStateRow.extras,
            )
            .where((PlayerStateRow.id == guild_id) & (PlayerStateRow.bot == bot_id))
            .first()
            .output(load_json=True, nested=True)
        )
        return cls(**player) if player else None



from __future__ import annotations

import pathlib
from typing import Final

import aiopath

from pylav.constants.config import CONFIG_DIR
from pylav.logging import getLogger

LOGGER = getLogger("PyLav.ManagedNode")

LAVALINK_DOWNLOAD_DIR = CONFIG_DIR / "lavalink"
_LAVALINK_DOWNLOAD_DIR = pathlib.Path(LAVALINK_DOWNLOAD_DIR)  # type: ignore
_LAVALINK_DOWNLOAD_DIR.mkdir(parents=True, exist_ok=True)
_LAVALINK_JAR_FILE_FORCED_SYNC = _LAVALINK_DOWNLOAD_DIR / "forced.jar"

LAVALINK_DOWNLOAD_DIR: aiopath.AsyncPath = aiopath.AsyncPath(LAVALINK_DOWNLOAD_DIR)
LAVALINK_JAR_FILE: aiopath.AsyncPath = LAVALINK_DOWNLOAD_DIR / "Lavalink.jar"
LAVALINK_JAR_FILE_FORCED: Final[aiopath.AsyncPath] = LAVALINK_DOWNLOAD_DIR / "forced.jar"
LAVALINK_APP_YML: Final[aiopath.AsyncPath] = LAVALINK_DOWNLOAD_DIR / "application.yml"
if USING_FORCED := _LAVALINK_JAR_FILE_FORCED_SYNC.exists():
    LOGGER.warning("%s found, disabling any JAR automated downloads", LAVALINK_JAR_FILE_FORCED)
    LAVALINK_JAR_FILE: aiopath.AsyncPath = LAVALINK_JAR_FILE_FORCED



from __future__ import annotations

from typing import TYPE_CHECKING

import discord

from pylav.events.base import PyLavEvent

if TYPE_CHECKING:
    from pylav.players.player import Player
    from pylav.players.tracks.obj import Track


class QueueEndEvent(PyLavEvent):
    """This event is dispatched when there are no more songs in the queue.

    Event can be listened to by adding a listener with the name `pylav_queue_end_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that has no more songs in queue.

    Parameters
    ----------
    player: :class:`Player`
        The player that has no more songs in queue.
    """

    __slots__ = ("player",)

    def __init__(self, player: Player) -> None:
        self.player = player


class QueueTrackPositionChangedEvent(PyLavEvent):
    """This event is dispatched when the position of a track is changed.

    Event can be listened to by adding a listener with the name `pylav_queue_track_position_changed_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player whose track position was changed.
    before: :class:`int`
        The position of the track before the change.
    after: :class:`int`
        The position of the track after the change.
    requester: :class:`discord.Member`
        The user who requested the change.
    track: :class:`Track`
        The track whose position was changed.

    Parameters
    ----------
    player: :class:`Player`
        The player whose track position was changed.
    before: :class:`int`
        The position of the track before the change.
    after: :class:`int`
        The position of the track after the change.
    requester: :class:`discord.Member`
        The user who requested the change.
    track: :class:`Track`
        The track whose position was changed.
    """

    __slots__ = ("player", "before", "after", "requester", "track")

    def __init__(self, player: Player, before: int, after: int, requester: discord.Member, track: Track) -> None:
        self.player = player
        self.requester = requester
        self.before = before
        self.after = after
        self.track = track


class QueueShuffledEvent(PyLavEvent):
    """This event is dispatched when the queue is shuffled.

    Event can be listened to by adding a listener with the name `pylav_queue_shuffled_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player whose queue was shuffled.
    requester: :class:`discord.Member`
        The user who requested the change.

    Parameters
    ----------
    player: :class:`Player`
        The player whose queue was shuffled.
    requester: :class:`discord.Member`
        The user who requested the change.
    """

    __slots__ = ("player", "requester")

    def __init__(self, player: Player, requester: discord.Member) -> None:
        self.player = player
        self.requester = requester


class QueueTracksRemovedEvent(PyLavEvent):
    """This event is dispatched when tracks are removed from the queue.

    Event can be listened to by adding a listener with the name `pylav_queue_tracks_removed_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player whose queue was shuffled.
    requester: :class:`discord.Member`
        The user who requested the change.
    tracks: :class:`list` of :class:`Track`
        The tracks that were removed from the queue.

    Parameters
    ----------
    player: :class:`Player`
        The player whose queue was shuffled.
    requester: :class:`discord.Member`
        The user who requested the change.
    tracks: :class:`list` of :class:`Track`
        The tracks that were removed from the queue.
    """

    __slots__ = ("player", "requester", "tracks")

    def __init__(self, player: Player, requester: discord.Member, tracks: list[Track]) -> None:
        self.player = player
        self.requester = requester
        self.tracks = tracks


class QueueTracksAddedEvent(PyLavEvent):
    """This event is dispatched when a track in added to the queue.

    Event can be listened to by adding a listener with the name `pylav_queue_tracks_added_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player whose queue had tracks added to it.
    requester: :class:`discord.Member`
        The user who requested the change.
    tracks: :class:`list` of :class:`Track`
        The tracks that were added to the queue.

    """

    __slots__ = ("player", "requester", "tracks")

    def __init__(self, player: Player, requester: discord.Member, tracks: list[Track]) -> None:
        self.player = player
        self.requester = requester
        self.tracks = tracks



from __future__ import annotations

from dataclasses import dataclass, field
from typing import NotRequired  # noqa


@dataclass(eq=True, slots=True, unsafe_hash=True, order=True, kw_only=True, frozen=True)
class LyricsExternal:
    """External IDs for a track."""

    isrc: str
    spotify_id: str


@dataclass(eq=True, slots=True, unsafe_hash=True, order=True, kw_only=True, frozen=True)
class LyricsMedia:
    """Media for a track."""

    preview: str | None
    artwork: str | None


@dataclass(eq=True, slots=True, unsafe_hash=True, order=True, kw_only=True, frozen=True)
class LyricsTrack:
    """Track information for a track."""

    title: str
    artist: str
    album: str
    duration: int
    explicit: bool
    external: LyricsExternal | dict
    media: LyricsMedia | dict

    def __post_init__(self):
        if isinstance(self.media, dict):
            object.__setattr__(self, "media", LyricsMedia(**self.media))
        if isinstance(self.external, dict):
            object.__setattr__(self, "external", LyricsExternal(**self.external))


@dataclass(eq=True, slots=True, unsafe_hash=True, order=True, kw_only=True, frozen=True)
class Task:
    """Task information for a query."""

    id: str
    status: str


@dataclass(eq=True, slots=True, unsafe_hash=True, order=True, kw_only=True, frozen=True)
class LyricsLine:
    """The timed lyric detaild for a section."""

    text: str
    start: int
    duration: int


@dataclass(eq=True, slots=True, unsafe_hash=True, order=True, kw_only=True, frozen=True)
class LyricsMetadata:
    """Lyrics metadata for a track."""

    text: str
    lines: list[LyricsLine]

    def __post_init__(self):
        temp = []
        for s in self.lines:
            if isinstance(s, LyricsLine) or (isinstance(s, dict) and (s := LyricsLine(**s))):
                temp.append(s)
        object.__setattr__(self, "lines", temp)


@dataclass(eq=True, slots=True, unsafe_hash=True, order=True, kw_only=True, frozen=True)
class Lyrics:
    """Lyrics for a track."""

    track: LyricsTrack | dict
    task: Task | None | dict
    lyrics: LyricsMetadata | None | dict = None
    provider: NotRequired[str] = field(
        repr=False, compare=False, hash=False, default="Flowery API - https://flowery.pw"
    )

    def __post_init__(self):
        if isinstance(self.track, dict):
            object.__setattr__(self, "track", LyricsTrack(**self.track))
        if isinstance(self.task, dict):
            object.__setattr__(self, "task", Task(**self.task))
        if isinstance(self.lyrics, dict):
            object.__setattr__(self, "lyrics", LyricsMetadata(**self.lyrics))


@dataclass(eq=True, slots=True, unsafe_hash=True, order=True, kw_only=True, frozen=True)
class Error:
    """Error information for a query."""

    error: str


@dataclass(eq=True, slots=True, unsafe_hash=True, order=True, kw_only=True, frozen=True)
class TrackList:
    """Track list information for a query."""

    tracks: list[LyricsTrack]

    def __post_init__(self):
        temp = []
        for s in self.tracks:
            if isinstance(s, LyricsTrack) or (isinstance(s, dict) and (s := LyricsTrack(**s))):
                temp.append(s)
        object.__setattr__(self, "tracks", temp)



from __future__ import annotations

from typing import TYPE_CHECKING

from packaging.version import Version

from pylav.constants.config import CONFIG_DIR
from pylav.storage.database.tables.aiohttp_cache import AioHttpCacheRow
from pylav.storage.database.tables.config import LibConfigRow
from pylav.storage.database.tables.equalizer import EqualizerRow
from pylav.storage.database.tables.m2m import TrackToPlaylists, TrackToQueries
from pylav.storage.database.tables.nodes import NodeRow, Sessions
from pylav.storage.database.tables.player_state import PlayerStateRow
from pylav.storage.database.tables.players import PlayerRow
from pylav.storage.database.tables.playlists import PlaylistRow
from pylav.storage.database.tables.queries import QueryRow
from pylav.storage.database.tables.tracks import TrackRow
from pylav.storage.database.tables.version import BotVersionRow
from pylav.storage.migrations.low_level.base import migrate_data, run_low_level_migrations
from pylav.storage.models.config import Config
from pylav.storage.models.version import BotVersion

if TYPE_CHECKING:
    from pylav.core.client import Client


class ConfigController:
    __slots__ = ("_client", "_config_folder")

    def __init__(self, client: Client) -> None:
        self._client = client
        self._config_folder = CONFIG_DIR

    async def initialize(self) -> None:
        data_to_migrate = await run_low_level_migrations(self)
        await self.create_tables()
        await migrate_data(data_to_migrate)

    @property
    def client(self) -> Client:
        return self._client

    # noinspection PyProtectedMember
    @staticmethod
    async def create_tables() -> None:
        await PlaylistRow.create_table(if_not_exists=True)
        await LibConfigRow.create_table(if_not_exists=True)
        await LibConfigRow.raw(
            f"CREATE UNIQUE INDEX IF NOT EXISTS unique_lib_config_bot_id "
            f"ON {LibConfigRow._meta.tablename} (bot, id)"
        )
        await EqualizerRow.create_table(if_not_exists=True)
        await PlayerStateRow.create_table(if_not_exists=True)
        await PlayerStateRow.raw(
            f"CREATE UNIQUE INDEX IF NOT EXISTS unique_player_state_bot_id "
            f"ON {PlayerStateRow._meta.tablename} (bot, id)"
        )
        await PlayerRow.create_table(if_not_exists=True)
        await PlayerRow.raw(
            f"CREATE UNIQUE INDEX IF NOT EXISTS unique_player_bot_id ON {PlayerRow._meta.tablename} (bot, id)"
        )
        await NodeRow.create_table(if_not_exists=True)
        await QueryRow.create_table(if_not_exists=True)
        await BotVersionRow.create_table(if_not_exists=True)
        await AioHttpCacheRow.create_table(if_not_exists=True)
        await TrackRow.create_table(if_not_exists=True)
        await TrackToPlaylists.create_table(if_not_exists=True)
        await TrackToQueries.create_table(if_not_exists=True)
        await Sessions.create_table(if_not_exists=True)
        await Sessions.raw(
            f"CREATE UNIQUE INDEX IF NOT EXISTS unique_node_bot_id ON {Sessions._meta.tablename} (bot, node)"
        )

    # noinspection PyProtectedMember
    async def reset_database(self) -> None:
        await PlaylistRow.raw(
            f"DROP TABLE IF EXISTS "
            f"{TrackToQueries._meta.tablename}, "
            f"{TrackToPlaylists._meta.tablename}, "
            f"{PlaylistRow._meta.tablename}, "
            f"{LibConfigRow._meta.tablename}, "
            f"{EqualizerRow._meta.tablename}, "
            f"{PlayerStateRow._meta.tablename}, "
            f"{PlayerRow._meta.tablename}, "
            f"{NodeRow._meta.tablename}, "
            f"{QueryRow._meta.tablename}, "
            f"{BotVersionRow._meta.tablename}, "
            f"{AioHttpCacheRow._meta.tablename}, "
            f"{TrackRow._meta.tablename}"
            ";"
        )
        await self.create_tables()

    def get_config(
        self,
    ) -> Config:
        return Config(id=1, bot=self.client.bot.user.id)

    def get_bot_db_version(self) -> BotVersion:
        return BotVersion(id=self._client.bot.user.id)

    async def update_bot_dv_version(self, version: str | Version) -> None:
        await self.get_bot_db_version().update_version(version)



from __future__ import annotations

import contextlib
import gzip
import io
import pathlib
import random
import sys
from collections import defaultdict
from collections.abc import Iterator
from dataclasses import dataclass

import aiohttp
import brotli  # type: ignore
import discord
import yaml
from dacite import from_dict

from pylav.compat import json
from pylav.constants.config import BROTLI_ENABLED, READ_CACHING_ENABLED
from pylav.constants.playlists import BUNDLED_PLAYLIST_IDS
from pylav.constants.regex import SQUARE_BRACKETS
from pylav.core.context import PyLavContext
from pylav.exceptions.playlist import InvalidPlaylistException
from pylav.helpers.singleton import SingletonCachedByKey
from pylav.logging import getLogger
from pylav.nodes.api.responses.track import Track
from pylav.storage.database.cache.decodators import maybe_cached
from pylav.storage.database.cache.model import CachedModel
from pylav.storage.database.tables.playlists import PlaylistRow
from pylav.storage.database.tables.tracks import TrackRow
from pylav.type_hints.bot import DISCORD_BOT_TYPE
from pylav.type_hints.dict_typing import JSON_DICT_TYPE

LOGGER = getLogger("PyLav.Database.Playlist")


try:
    from redbot.core.i18n import Translator  # type: ignore

    _ = Translator("PyLav", pathlib.Path(__file__))
except ImportError:
    Translator = None

    def _(string: str) -> str:
        return string


@dataclass(eq=True, slots=True, unsafe_hash=True, order=True, kw_only=True, frozen=True)
class Playlist(CachedModel, metaclass=SingletonCachedByKey):
    id: int

    def get_cache_key(self) -> str:
        return f"{self.id}"

    @maybe_cached
    async def exists(self) -> bool:
        """Check if the config exists.

        Returns
        -------
        bool
            Whether the config exists.
        """

        return await PlaylistRow.exists().where(PlaylistRow.id == self.id)

    @maybe_cached
    async def fetch_all(self) -> JSON_DICT_TYPE:
        """Fetch all playlists from the database.

        Returns
        -------
        dict
            The playlists.
        """
        data = (
            await PlaylistRow.select(
                PlaylistRow.id,
                PlaylistRow.name,
                PlaylistRow.tracks(TrackRow.encoded, TrackRow.info, TrackRow.pluginInfo, load_json=True),
                PlaylistRow.scope,
                PlaylistRow.author,
                PlaylistRow.url,
            )
            .where(PlaylistRow.id == self.id)
            .first()
            .output(load_json=True, nested=True)
        )
        return data or {
            "id": self.id,
            "name": PlaylistRow.name.default,
            "tracks": [],
            "scope": PlaylistRow.scope.default,
            "author": PlaylistRow.author.default,
            "url": PlaylistRow.url.default,
        }

    @maybe_cached
    async def fetch_scope(self) -> int | None:
        """Fetch the scope of the playlist.

        Returns
        -------
        int
            The scope of the playlist.
        """
        response = (
            await PlaylistRow.select(PlaylistRow.scope)
            .where(PlaylistRow.id == self.id)
            .first()
            .output(load_json=True, nested=True)
        )
        return response["scope"] if response else PlaylistRow.scope.default

    async def update_scope(self, scope: int):
        """Update the scope of the playlist.

        Parameters
        ----------
        scope : int
            The new scope of the playlist.
        """
        await PlaylistRow.insert(PlaylistRow(id=self.id, scope=scope)).on_conflict(
            action="DO UPDATE", target=PlaylistRow.id, values=[PlaylistRow.scope]
        )
        await self.update_cache((self.fetch_scope, scope), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_author(self) -> int | None:
        """Fetch the author of the playlist.

        Returns
        -------
        int
            The author of the playlist.
        """
        response = (
            await PlaylistRow.select(PlaylistRow.author)
            .where(PlaylistRow.id == self.id)
            .first()
            .output(load_json=True, nested=True)
        )
        return response["author"] if response else PlaylistRow.author.default

    async def update_author(self, author: int) -> None:
        """Update the author of the playlist.

        Parameters
        ----------
        author : int
            The new author of the playlist.
        """
        await PlaylistRow.insert(PlaylistRow(id=self.id, author=author)).on_conflict(
            action="DO UPDATE", target=PlaylistRow.id, values=[PlaylistRow.author]
        )
        await self.update_cache((self.fetch_author, author), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_name(self) -> str | None:
        """Fetch the name of the playlist.

        Returns
        -------
        str
            The name of the playlist.
        """
        response = (
            await PlaylistRow.select(PlaylistRow.name)
            .where(PlaylistRow.id == self.id)
            .first()
            .output(load_json=True, nested=True)
        )
        return response["name"] if response else PlaylistRow.name.default

    async def update_name(self, name: str) -> None:
        """Update the name of the playlist.

        Parameters
        ----------
        name : str
            The new name of the playlist.
        """
        await PlaylistRow.insert(PlaylistRow(id=self.id, name=name)).on_conflict(
            action="DO UPDATE", target=PlaylistRow.id, values=[PlaylistRow.name]
        )
        await self.update_cache((self.fetch_name, name), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_url(self) -> str | None:
        """Fetch the url of the playlist.

        Returns
        -------
        str
            The url of the playlist.
        """
        response = (
            await PlaylistRow.select(PlaylistRow.url)
            .where(PlaylistRow.id == self.id)
            .first()
            .output(load_json=True, nested=True)
        )
        return response["url"] if response else PlaylistRow.url.default

    async def update_url(self, url: str) -> None:
        """Update the url of the playlist.

        Parameters
        ----------
        url : str
            The new url of the playlist.
        """
        await PlaylistRow.insert(PlaylistRow(id=self.id, url=url)).on_conflict(
            action="DO UPDATE", target=PlaylistRow.id, values=[PlaylistRow.url]
        )
        await self.update_cache((self.fetch_url, url), (self.exists, True))
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def fetch_tracks(self) -> list[str | JSON_DICT_TYPE]:
        """Fetch the tracks of the playlist.

        Returns
        -------
        list[str]
            The tracks of the playlist.
        """
        response = (
            await PlaylistRow.select(
                PlaylistRow.tracks(TrackRow.encoded, TrackRow.info, TrackRow.pluginInfo, load_json=True)
            )
            .where(PlaylistRow.id == self.id)
            .first()
            .output(load_json=True, nested=True)
        )
        data = response["tracks"] if response else []
        return data

    async def update_tracks(self, tracks: list[str | JSON_DICT_TYPE | Track]) -> None:
        """Update the tracks of the playlist.

        Parameters
        ----------
        tracks : list[str]
            The new tracks of the playlist.
        """
        playlist_row = await PlaylistRow.objects().get_or_create(PlaylistRow.id == self.id)
        try:
            old_tracks = await playlist_row.get_m2m(PlaylistRow.tracks)
        except ValueError:
            old_tracks = []
        new_tracks = []
        # TODO: Optimize this, after https://github.com/piccolo-orm/piccolo/discussions/683 is answered or fixed
        _temp = defaultdict(list)
        for x in tracks:
            _temp[type(x)].append(x)
        for entry_type, entry_list in _temp.items():
            if entry_type == str:
                for track_object in await self.client.decode_tracks(entry_list, raise_on_failure=False):
                    new_tracks.append(await TrackRow.get_or_create(track_object))
            elif entry_type == dict:
                for track_object in entry_list:
                    new_tracks.append(await TrackRow.get_or_create(from_dict(data_class=Track, data=track_object)))
            else:
                for track_object in entry_list:
                    new_tracks.append(await TrackRow.get_or_create(track_object))

        if old_tracks:
            await playlist_row.remove_m2m(*old_tracks, m2m=PlaylistRow.tracks)
        if new_tracks:
            await playlist_row.add_m2m(*new_tracks, m2m=PlaylistRow.tracks)

        await self.invalidate_cache(self.fetch_tracks, self.fetch_first)
        await self.update_cache(
            (self.exists, True),
            (self.size, len(tracks)),
        )
        await self.invalidate_cache(self.fetch_all)

    @maybe_cached
    async def size(self) -> int:
        """Count the tracks of the playlist.

        Returns
        -------
        int
            The number of tracks in the playlist.
        """
        tracks = await self.fetch_tracks()
        return len(tracks) if tracks else 0

    async def add_track(self, tracks: list[str | Track | JSON_DICT_TYPE]) -> None:
        """Add a track to the playlist.

        Parameters
        ----------
        tracks : list[str | Track]
            The tracks to add.
        """
        playlist_row = await PlaylistRow.objects().get_or_create(PlaylistRow.id == self.id)
        new_tracks = []
        # TODO: Optimize this, after https://github.com/piccolo-orm/piccolo/discussions/683 is answered or fixed
        _temp = defaultdict(list)
        for x in tracks:
            _temp[type(x)].append(x)
        for entry_type, entry_list in _temp.items():
            if entry_type == str:
                for track_object in await self.client.decode_tracks(entry_list, raise_on_failure=False):
                    new_tracks.append(await TrackRow.get_or_create(track_object))
            elif entry_type == dict:
                for track_object in entry_list:
                    new_tracks.append(await TrackRow.get_or_create(from_dict(data_class=Track, data=track_object)))
            else:
                for track_object in entry_list:
                    new_tracks.append(await TrackRow.get_or_create(track_object))
        if new_tracks:
            await playlist_row.add_m2m(*new_tracks, m2m=PlaylistRow.tracks)
        await self.invalidate_cache(self.fetch_tracks, self.fetch_all, self.size, self.fetch_first, self.exists)

    async def bulk_remove_tracks(self, tracks: list[str]) -> None:
        """Remove disc jockey users from the player.

        Parameters
        ----------
        tracks : list[str]
            The track to remove
        """
        if not tracks:
            return
        playlist = await PlaylistRow.objects().where(PlaylistRow.id == self.id).first()
        tracks = await TrackRow.objects().where(TrackRow.encoded.is_in(tracks))
        if tracks:
            await playlist.remove_m2m(*tracks, m2m=PlaylistRow.tracks)
        await self.invalidate_cache(self.fetch_tracks, self.fetch_all, self.size, self.fetch_first, self.exists)

    async def remove_track(self, track: str) -> None:
        """Remove a track from the playlist.

        Parameters
        ----------
        track : str
            The track to remove
        """
        return await self.bulk_remove_tracks([track])

    async def remove_all_tracks(self) -> None:
        """Remove all tracks from the playlist."""
        playlist = await PlaylistRow.objects().where(PlaylistRow.id == self.id).first()
        try:
            tracks = await playlist.get_m2m(PlaylistRow.tracks)
        except ValueError:
            tracks = []
        if tracks:
            await playlist.remove_m2m(*tracks, m2m=PlaylistRow.tracks)
        await self.update_cache((self.fetch_tracks, []), (self.size, 0), (self.exists, True), (self.fetch_first, None))
        await self.invalidate_cache(self.fetch_all)

    async def delete(self) -> None:
        """Delete the playlist from the database"""
        await PlaylistRow.delete().where(PlaylistRow.id == self.id)
        await self.invalidate_cache()

    async def can_manage(self, bot: DISCORD_BOT_TYPE, requester: discord.abc.User) -> bool:  # noqa
        """Check if the requester can manage the playlist.

        Parameters
        ----------
        bot : DISCORD_BOT_TYPE
            The bot instance.
        requester : discord.abc.User
            The requester.

        Returns
        -------
        bool
            Whether the requester can manage the playlist.
        """
        if self.id in BUNDLED_PLAYLIST_IDS:
            return False
        if requester.id in ((ids := getattr(bot, "owner_ids")) or ()) or requester.id == bot.owner_id:  # noqa
            return True
        if await self.fetch_scope() == bot.user.id:
            return False
        return await self.fetch_author() == requester.id

    async def get_scope_name(self, bot: DISCORD_BOT_TYPE, mention: bool = True, guild: discord.Guild = None) -> str:
        """Get the name of the scope of the playlist.

        Parameters
        ----------
        bot : DISCORD_BOT_TYPE
            The bot instance.
        mention : bool
            Whether to add a mention if it is mentionable.
        guild : discord.Guild
            The guild to get the scope name for.

        Returns
        -------
        str
            The name of the scope of the playlist.
        """
        original_scope = await self.fetch_scope()
        if bot.user.id == original_scope:
            return _("(Global) {user_name_variable_do_not_translate}").format(
                user_name_variable_do_not_translate=bot.user.mention if mention else bot.user
            )
        elif guild_ := bot.get_guild(original_scope):
            if guild_:
                guild = guild_
            return _("(Server) {guild_name_variable_do_not_translate}").format(
                guild_name_variable_do_not_translate=guild.name
            )
        elif guild and (channel := guild.get_channel_or_thread(original_scope)):
            return _("(Channel) {channel_name_variable_do_not_translate}").format(
                channel_name_variable_do_not_translate=channel.mention if mention else channel.name
            )
        elif (
            (guild := guild_ or guild)
            and (guild and (scope := guild.get_member(original_scope)))  # noqa
            or (scope := bot.get_user(original_scope))
        ):
            return _("(User) {user_name_variable_do_not_translate}").format(
                user_name_variable_do_not_translate=scope.mention if mention else scope
            )
        else:
            return _("(Invalid) {scope_name_variable_do_not_translate}").format(
                scope_name_variable_do_not_translate=original_scope
            )

    async def get_author_name(self, bot: DISCORD_BOT_TYPE, mention: bool = True) -> str | None:
        """Get the name of the author of the playlist.

        Parameters
        ----------
        bot : DISCORD_BOT_TYPE
            The bot instance.
        mention : bool
            Whether to add a mention if it is mentionable.

        Returns
        -------
        str | None
            The name of the author of the playlist.
        """
        author = await self.fetch_author()
        if user := bot.get_user(author):
            return f"{user.mention}" if mention else f"{user}"
        return f"{author}"

    async def get_name_formatted(self, with_url: bool = True, escape: bool = True) -> str:
        """Get the name of the playlist formatted.

        Parameters
        ----------
        with_url : bool
            Whether to include the url in the name.
        escape: bool
            Whether to markdown escape the response
        Returns
        -------
        str
            The formatted name.
        """
        name = SQUARE_BRACKETS.sub("", await self.fetch_name()).strip()
        if with_url:
            url = await self.fetch_url()
            if url and url.startswith("http"):
                return f"**[{discord.utils.escape_markdown(name) if escape else name}]({url})**"
        return f"**{discord.utils.escape_markdown(name) if escape else name}**"

    @contextlib.asynccontextmanager
    async def to_yaml(self, guild: discord.Guild) -> Iterator[tuple[io.BytesIO, str | None]]:
        """Serialize the playlist to a YAML file.

        yields a tuple of (io.BytesIO, bool) where the bool is whether the playlist file was compressed using Gzip

        Parameters
        ----------
        guild : discord.Guild
            The guild where the yaml will be sent to.

        Yields
        ------
        tuple[io.BytesIO, str | None]
            The YAML file and the compression type.
        """
        data = await self.fetch_all()
        name = data["name"]
        compression = None
        with io.BytesIO() as bio:
            yaml.safe_dump(data, bio, default_flow_style=False, sort_keys=False, encoding="utf-8")
            bio.seek(0)
            LOGGER.debug("SIZE UNCOMPRESSED playlist (%s): %s", name, sys.getsizeof(bio))
            if sys.getsizeof(bio) > guild.filesize_limit:
                with io.BytesIO() as cbio:
                    if BROTLI_ENABLED:
                        compression = "brotli"
                        cbio.write(brotli.compress(yaml.dump(data, encoding="utf-8")))
                    else:
                        compression = "gzip"
                        with gzip.GzipFile(fileobj=cbio, mode="wb", compresslevel=9) as gzip_file:
                            yaml.safe_dump(data, gzip_file, default_flow_style=False, sort_keys=False, encoding="utf-8")
                    cbio.seek(0)
                    LOGGER.debug("SIZE COMPRESSED playlist [%s] (%s): %s", compression, name, sys.getsizeof(cbio))
                    yield cbio, compression
                    return
            yield bio, compression

    async def bulk_update(
        self, scope: int, name: str, author: int, url: str | None, tracks: list[str | JSON_DICT_TYPE | Track]
    ) -> None:
        """Bulk update the playlist."""
        defaults = {
            PlaylistRow.name: name,
            PlaylistRow.author: author,
            PlaylistRow.scope: scope,
            PlaylistRow.url: url,
        }

        playlist_row = await PlaylistRow.objects().get_or_create(PlaylistRow.id == self.id, defaults)
        # noinspection PyProtectedMember
        if not playlist_row._was_created:
            await PlaylistRow.update(defaults).where(PlaylistRow.id == self.id)
        try:
            old_tracks = await playlist_row.get_m2m(PlaylistRow.tracks)
        except ValueError:
            old_tracks = []
        new_tracks = []
        # TODO: Optimize this, after https://github.com/piccolo-orm/piccolo/discussions/683 is answered or fixed
        _temp = defaultdict(list)
        for x in tracks:
            _temp[type(x)].append(x)
        for entry_type, entry_list in _temp.items():
            if entry_type == str:
                for track_object in await self.client.decode_tracks(entry_list, raise_on_failure=False):
                    new_tracks.append(await TrackRow.get_or_create(track_object))
            elif entry_type == dict:
                for track_object in entry_list:
                    new_tracks.append(await TrackRow.get_or_create(from_dict(data_class=Track, data=track_object)))
            else:
                for track_object in entry_list:
                    new_tracks.append(await TrackRow.get_or_create(track_object))
        if old_tracks:
            await playlist_row.remove_m2m(*old_tracks, m2m=PlaylistRow.tracks)
        if new_tracks:
            await playlist_row.add_m2m(*new_tracks, m2m=PlaylistRow.tracks)
        await self.invalidate_cache()

    @classmethod
    async def from_yaml(cls, context: PyLavContext, scope: int, url: str) -> Playlist:
        """Deserialize a playlist from a YAML file.

        Parameters
        ----------
        context : PyLavContext
            The context.
        scope : int
            The scope of the playlist.
        url : str
            The url of the playlist.

        Returns
        -------
        Playlist
            The playlist.
        """
        try:
            async with aiohttp.ClientSession(auto_decompress=False, json_serialize=json.dumps) as session:
                async with session.get(url) as response:
                    data = await response.read()
                    if ".gz.pylav" in url:
                        data = gzip.decompress(data)
                    elif ".br.pylav" in url:
                        data = brotli.decompress(data)
                    data = yaml.safe_load(data)
        except Exception as e:
            raise InvalidPlaylistException(f"Invalid playlist file - {e}") from e
        playlist = cls(
            id=context.message.id,
        )
        await playlist.bulk_update(
            scope=scope, name=data["name"], url=data["url"], tracks=data["tracks"], author=context.author.id
        )
        return playlist

    async def fetch_index(self, index: int) -> JSON_DICT_TYPE | None:
        """Get the track at the index.

        Parameters
        ----------
        index: int
            The index of the track

        Returns
        -------
        str
            The track at the index
        """
        if READ_CACHING_ENABLED:
            tracks = await self.fetch_tracks()
            return tracks[index] if index < len(tracks) else None
        else:
            tracks = await self.fetch_tracks()
            if tracks and len(tracks) > index:
                return tracks[index]

    @maybe_cached
    async def fetch_first(self) -> JSON_DICT_TYPE | None:
        """Get the first track.

        Returns
        -------
        str
            The first track
        """
        return await self.fetch_index(0)

    async def fetch_random(self) -> JSON_DICT_TYPE | None:
        """Get a random track.

        Returns
        -------
        str
            A random track
        """

        return await self.fetch_index(random.randint(0, await self.size()))



from __future__ import annotations

from collections.abc import AsyncIterable
from typing import Any

from aiohttp_client_cache import BaseCache, CacheBackend, ResponseOrKey
from aiohttp_client_cache.signatures import extend_init_signature

from pylav.storage.database.tables.aiohttp_cache import AioHttpCacheRow


def postgres_template() -> None:
    pass


@extend_init_signature(CacheBackend, postgres_template)
class PostgresCacheBackend(CacheBackend):
    """Wrapper for higher-level cache operations.
    In most cases, the only thing you need to specify here is which storage class(es) to use"""

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.redirects = PostgresStorage(**kwargs)
        self.responses = PostgresStorage(**kwargs)


class PostgresStorage(BaseCache):
    """interface for lower-level backend storage operations"""

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)

    async def contains(self, key: str) -> bool:
        """Check if a key is stored in the cache"""
        return await AioHttpCacheRow.exists().where(AioHttpCacheRow.key == key)

    async def clear(self) -> None:
        """Delete all items from the cache"""
        await AioHttpCacheRow.raw("TRUNCATE TABLE aiohttp_client_cache")

    async def delete(self, key: str) -> None:
        """Delete an item from the cache"""
        await AioHttpCacheRow.delete().where(AioHttpCacheRow.key == key)

    async def keys(self) -> AsyncIterable[str]:
        """Get all keys stored in the cache"""
        for entry in await AioHttpCacheRow.select(AioHttpCacheRow.key).output(load_json=True, nested=True):
            yield entry["key"]

    async def read(self, key: str) -> ResponseOrKey:
        """Read an item from the cache"""
        response = (
            await AioHttpCacheRow.select(AioHttpCacheRow.value)
            .where(AioHttpCacheRow.key == key)
            .first()
            .output(load_json=True, nested=True)
        )
        return self.deserialize(response["value"]) if response else None

    async def size(self) -> int:
        """Get the number of items in the cache"""
        return await AioHttpCacheRow.count()

    def values(self) -> AsyncIterable[ResponseOrKey]:
        """Get all values stored in the cache"""
        return self._values()

    async def _values(self) -> AsyncIterable[ResponseOrKey]:
        for entry in await AioHttpCacheRow.select(AioHttpCacheRow.value).output(load_json=True, nested=True):
            yield self.deserialize(entry["value"])

    async def write(self, key: str, item: ResponseOrKey) -> None:
        """Write an item to the cache"""
        await AioHttpCacheRow.insert(AioHttpCacheRow(key=key, value=self.serialize(item))).on_conflict(
            action="DO NOTHING", target=AioHttpCacheRow.key
        )

    async def bulk_delete(self, keys: set[str]) -> None:
        """Delete multiple items from the cache"""
        await AioHttpCacheRow.delete().where(AioHttpCacheRow.key.is_in(list(keys)))



from __future__ import annotations

import collections
from typing import TYPE_CHECKING, Literal

import discord

from pylav.events.base import PyLavEvent
from pylav.nodes.node import Node

if TYPE_CHECKING:
    from pylav.players.filters import (
        ChannelMix,
        Distortion,
        Echo,
        Equalizer,
        Karaoke,
        LowPass,
        Rotation,
        Timescale,
        Tremolo,
        Vibrato,
        Volume,
    )
    from pylav.players.player import Player
    from pylav.players.tracks.obj import Track


class PlayerUpdateEvent(PyLavEvent):
    """This event is dispatched when the player's progress changes.

    Event can be listened to by adding a listener with the name `pylav_player_update_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that was updated.
    position: :class:`float`
        The position of the player.
    timestamp: :class:`float`
        The timestamp of the event.
    """

    __slots__ = ("player", "position", "timestamp")

    def __init__(self, player: Player, position: float, timestamp: float) -> None:
        self.player = player
        self.position = position
        self.timestamp = timestamp


class PlayerPausedEvent(PyLavEvent):
    """This event is dispatched when a player is paused.

    Event can be listened to by adding a listener with the name `pylav_player_paused_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that was paused.
    requester: :class:`discord.Member`
        The member that requested the pause.
    """

    __slots__ = ("player", "requester")

    def __init__(self, player: Player, requester: discord.Member) -> None:
        self.player = player
        self.requester = requester


class PlayerStoppedEvent(PyLavEvent):
    """This event is dispatched when a player is stopped.

    Event can be listened to by adding a listener with the name `pylav_player_stopped_event`.


    Attributes
    ----------
    player: :class:`Player`
        The player that was stopped.
    requester: :class:`discord.Member`
        The member that requested the stop.
    """

    __slots__ = ("player", "requester")

    def __init__(self, player: Player, requester: discord.Member) -> None:
        self.player = player
        self.requester = requester


class PlayerResumedEvent(PyLavEvent):
    """This event is dispatched when a player is resumed.

    Event can be listened to by adding a listener with the name `pylav_player_resumed_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that was resumed.
    requester: :class:`discord.Member`
        The member that requested the resume.

    """

    __slots__ = ("player", "requester")

    def __init__(self, player: Player, requester: discord.Member) -> None:
        self.player = player
        self.requester = requester


class PlayerRestoredEvent(PyLavEvent):
    """This event is dispatched when a player is restored.

    Event can be listened to by adding a listener with the name `pylav_player_restored_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that was restored.
    requester: :class:`discord.Member`
        The member that requested the restore.
    """

    __slots__ = ("player", "requester")

    def __init__(self, player: Player, requester: discord.abc.User) -> None:
        self.player = player
        self.requester = requester


class PlayerMovedEvent(PyLavEvent):
    """This event is dispatched when the player is moved.

    Event can be listened to by adding a listener with the name `pylav_player_moved_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that was moved.
    requester: :class:`discord.Member`
        The member that requested the move.
    before: :class:`discord.channel.VocalGuildChannel`
        The channel the player was in before the move.
    after: :class:`discord.channel.VocalGuildChannel`
        The channel the player is in after the move.
    """

    __slots__ = ("player", "requester", "before", "after")

    def __init__(
        self,
        player: Player,
        requester: discord.Member,
        before: discord.channel.VocalGuildChannel,
        after: discord.channel.VocalGuildChannel,
    ) -> None:
        self.player = player
        self.requester = requester
        self.before = before
        self.after = after


class PlayerDisconnectedEvent(PyLavEvent):
    """This event is dispatched when the player is disconnected.

    Event can be listened to by adding a listener with the name `pylav_player_disconnected_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that was disconnected.
    requester: :class:`discord.Member`
        The member that requested the disconnect.
    current_track: :class:`Track` | :class:`None`
        The current track that was playing when the player was disconnected.
    position: :class:`float`
        The position of the current track when the player was disconnected.
    queue: :class:`collections.deque[Track]`
        The queue of the player when the player was disconnected.
    """

    __slots__ = ("player", "requester", "current_track", "position", "queue")

    def __init__(self, player: Player, requester: discord.Member) -> None:
        self.player = player
        self.requester = requester
        self.current_track: Track | None = player.current
        self.position: float = player.estimated_position
        self.queue: collections.deque[Track] = player.queue.raw_queue


class PlayerConnectedEvent(PyLavEvent):
    """This event is dispatched when the player is connected.

    Event can be listened to by adding a listener with the name `pylav_player_connected_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that was connected.
    requester: :class:`discord.Member` | :class:`None`
        The member that requested the connect.
    """

    __slots__ = ("player", "requester")

    def __init__(self, player: Player, requester: discord.Member | None) -> None:
        self.player = player
        self.requester = requester


class PlayerVolumeChangedEvent(PyLavEvent):
    """This event is dispatched when the player has its volume changed.

    Event can be listened to by adding a listener with the name `pylav_player_volume_changed_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that had its volume changed.
    requester: :class:`discord.Member`
        The member that requested the volume change.
    before: :class:`float`
        The volume before the change.
    after: :class:`float`
        The volume after the change.
    """

    __slots__ = ("player", "requester", "before", "after")

    def __init__(self, player: Player, requester: discord.Member, before: int, after: int) -> None:
        self.player = player
        self.requester = requester
        self.before = before
        self.after = after


class PlayerRepeatEvent(PyLavEvent):
    """This event is dispatched when the player repeat config changed.

    Event can be listened to by adding a listener with the name `pylav_player_repeat_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that had its repeat config changed.
    requester: :class:`discord.Member`
        The member that requested the repeat change.
    type: :class:`str`
        The type of operation that was performed, "current" or "queue"
    queue_before: :class:`bool`
        The queue repeat state before the operation.
    queue_after: :class:`bool`
        The queue repeat state after the operation.
    current_before: :class:`bool`
        The current track repeat state before the operation.
    current_after: :class:`bool`
        The current track repeat state after the operation.
    """

    __slots__ = ("player", "requester", "type", "queue_before", "queue_after", "current_before", "current_after")

    def __init__(
        self,
        player: Player,
        requester: discord.Member,
        op_type: Literal["current", "queue"],
        queue_before: bool,
        queue_after: bool,
        current_before: bool,
        current_after: bool,
    ) -> None:
        self.player = player
        self.requester = requester
        self.queue_before = queue_before
        self.queue_after = queue_after
        self.current_before = current_before
        self.current_after = current_after
        self.type = op_type


class FiltersAppliedEvent(PyLavEvent):
    """This event is dispatched when filters are applied.

    Event can be listened to by adding a listener with the name `filters_applied_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that had filters applied.
    requester: :class:`discord.Member`
        The member that requested the filters to be applied.
    volume: :class:`Volume` | :class:`None`
        The volume filter.
    equalizer: :class:`Equalizer` | :class:`None`
        The equalizer filter.
    karaoke: :class:`Karaoke` | :class:`None`
        The karaoke filter.
    timescale: :class:`Timescale` | :class:`None`
        The timescale filter.
    tremolo: :class:`Tremolo` | :class:`None`
        The tremolo filter.
    vibrato: :class:`Vibrato` | :class:`None`
        The vibrato filter.
    rotation: :class:`Rotation` | :class:`None`
        The rotation filter.
    distortion: :class:`Distortion` | :class:`None`
        The distortion filter.
    low_pass: :class:`LowPass` | :class:`None`
        The low pass filter.
    channel_mix: :class:`ChannelMix` | :class:`None`
        The channel mix filter.
    echo: :class:`Echo` | :class:`None`
        The echo filter.
    """

    __slots__ = (
        "player",
        "requester",
        "volume",
        "equalizer",
        "karaoke",
        "timescale",
        "tremolo",
        "vibrato",
        "rotation",
        "distortion",
        "low_pass",
        "channel_mix",
        "echo",
    )

    def __init__(
        self,
        player: Player,
        requester: discord.Member,
        node: Node,
        volume: Volume | None = None,
        equalizer: Equalizer | None = None,
        karaoke: Karaoke | None = None,
        timescale: Timescale | None = None,
        tremolo: Tremolo | None = None,
        vibrato: Vibrato | None = None,
        rotation: Rotation | None = None,
        distortion: Distortion | None = None,
        low_pass: LowPass | None = None,
        channel_mix: ChannelMix | None = None,
        echo: Echo | None = None,
    ) -> None:
        self.player = player
        self.requester = requester
        self.volume = volume
        self.equalizer = equalizer
        self.karaoke = karaoke
        self.timescale = timescale
        self.tremolo = tremolo
        self.vibrato = vibrato
        self.rotation = rotation
        self.distortion = distortion
        self.low_pass = low_pass
        self.channel_mix = channel_mix
        self.node = node
        self.echo = echo


class QuickPlayEvent(PyLavEvent):
    """This event is dispatched when a track is played with higher priority than the current track.

    Event can be listened to by adding a listener with the name `quick_play_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that had a track played.
    requester: :class:`discord.Member`
        The member that requested the track to be played.
    track: :class:`Track`
        The track that was played.
    """

    __slots__ = ("player", "requester", "track")

    def __init__(self, player: Player, requester: discord.Member, track: Track) -> None:
        self.player = player
        self.requester = requester
        self.track = track


class PlayerAutoDisconnectedEmptyQueueEvent(PyLavEvent):
    """This event is dispatched when the player is auto disconnected due to an empty queue.

    Event can be listened to by adding a listener with the name `player_auto_disconnected_empty_queue_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that was auto disconnected.
    requester: :class:`discord.Member`
        The member that requested the player to be auto disconnected.
    current_track: :class:`Track` | :class:`None`
        The current track that was playing.
    position: :class:`float`
        The position of the current track.
    queue: :class:`collections.deque[Track]`
        The queue of the player when it was auto disconnected.

    """

    __slots__ = ("player", "requester", "current_track", "position", "queue")

    def __init__(self, player: Player) -> None:
        self.player = player
        self.requester = player.guild.me
        self.current_track: Track | None = player.current
        self.position: float = player.estimated_position
        self.queue: collections.deque[Track] = player.queue.raw_queue


class PlayerAutoDisconnectedAloneEvent(PyLavEvent):
    """This event is dispatched when the player is auto disconnected due to it being alone in the voice channel.

    Event can be listened to by adding a listener with the name `player_auto_disconnected_alone_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that was auto disconnected.
    requester: :class:`discord.Member`
        The member that requested the player to be auto disconnected.
    current_track: :class:`Track` | :class:`None`
        The current track that was playing.
    position: :class:`float`
        The position of the current track.
    queue: :class:`collections.deque[Track]`
        The queue of the player when it was auto disconnected.
    """

    __slots__ = ("player", "requester", "current_track", "position", "queue")

    def __init__(self, player: Player) -> None:
        self.player = player
        self.requester = player.guild.me
        self.current_track: Track | None = player.current
        self.position: float = player.estimated_position
        self.queue: collections.deque[Track] = player.queue.raw_queue


class PlayerAutoPausedEvent(PyLavEvent):
    """This event is dispatched when the player is auto paused.

    Event can be listened to by adding a listener with the name `player_auto_paused_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that was auto paused.
    requester: :class:`discord.Member`
        The member that requested the player to be auto paused.
    """

    __slots__ = ("player", "requester")

    def __init__(self, player: Player) -> None:
        self.player = player
        self.requester = player.guild.me


class PlayerAutoResumedEvent(PyLavEvent):
    """This event is dispatched when the player is auto resumed.

    Event can be listened to by adding a listener with the name `player_auto_resumed_event`.

    Attributes
    ----------
    player: :class:`Player`
        The player that was auto resumed.
    requester: :class:`discord.Member`
        The member that requested the player to be auto resumed.
    """

    __slots__ = ("player", "requester")

    def __init__(self, player: Player) -> None:
        self.player = player
        self.requester = player.guild.me



from __future__ import annotations

import typing

import importlib_metadata
from packaging.version import Version, parse

from pylav.__version__ import __version__ as __version__

try:
    __PATH = None if (__PATHS := importlib_metadata.files("Py-Lav")) is None else next(iter(__PATHS), None)
except importlib_metadata.PackageNotFoundError:
    __PATH = None
VERSION: Version = typing.cast(Version, parse(__version__))
LOCATION = __PATH.locate() if __PATH else None


__all__ = (
    "__version__",
    "VERSION",
    "LOCATION",
)



from __future__ import annotations

import asyncio
import contextlib
from pathlib import Path
from typing import Any

import discord
from redbot.core.i18n import Translator

from pylav.constants.playlists import BUNDLED_PLAYLIST_IDS
from pylav.core.context import PyLavContext
from pylav.extension.red.ui.buttons.generic import CloseButton, DoneButton, NavigateButton, RefreshButton
from pylav.extension.red.ui.buttons.playlist import (
    EnqueuePlaylistButton,
    PlaylistClearButton,
    PlaylistDedupeButton,
    PlaylistDeleteButton,
    PlaylistDownloadButton,
    PlaylistInfoButton,
    PlaylistQueueButton,
    PlaylistUpdateButton,
    PlaylistUpsertButton,
)
from pylav.extension.red.ui.menus.generic import BaseMenu
from pylav.extension.red.ui.modals.generic import PromptForInput
from pylav.extension.red.ui.selectors.playlist import PlaylistPlaySelector, PlaylistSelectSelector
from pylav.extension.red.ui.sources.playlist import PlaylistPickerSource
from pylav.helpers import emojis
from pylav.helpers.format.strings import shorten_string
from pylav.storage.models.playlist import Playlist
from pylav.type_hints.bot import DISCORD_BOT_TYPE, DISCORD_COG_TYPE, DISCORD_CONTEXT_TYPE, DISCORD_INTERACTION_TYPE

_ = Translator("PyLav", Path(__file__))


class PlaylistPickerMenu(BaseMenu):
    _source: PlaylistPickerSource
    result: Playlist

    def __init__(
        self,
        cog: DISCORD_COG_TYPE,
        bot: DISCORD_BOT_TYPE,
        source: PlaylistPickerSource,
        selector_text: str,
        selector_cls: type[PlaylistPlaySelector] | type[PlaylistSelectSelector],  # noqa
        original_author: discord.abc.User,
        *,
        clear_buttons_after: bool = False,
        delete_after_timeout: bool = True,
        timeout: int = 120,
        message: discord.Message = None,
        starting_page: int = 0,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            cog,
            bot,
            source,
            clear_buttons_after=clear_buttons_after,
            delete_after_timeout=delete_after_timeout,
            timeout=timeout,
            message=message,
            starting_page=starting_page,
            **kwargs,
        )
        self.result: Playlist | None = None
        self.selector_cls = selector_cls
        self.selector_text = shorten_string(max_length=100, string=selector_text)
        self.forward_button = NavigateButton(
            style=discord.ButtonStyle.grey,
            emoji="\N{BLACK RIGHT-POINTING TRIANGLE}\N{VARIATION SELECTOR-16}",
            direction=1,
            row=4,
            cog=cog,
        )
        self.backward_button = NavigateButton(
            style=discord.ButtonStyle.grey,
            emoji="\N{BLACK LEFT-POINTING TRIANGLE}\N{VARIATION SELECTOR-16}",
            direction=-1,
            row=4,
            cog=cog,
        )
        self.first_button = NavigateButton(
            style=discord.ButtonStyle.grey,
            emoji="\N{BLACK LEFT-POINTING DOUBLE TRIANGLE}",
            direction=0,
            row=4,
            cog=cog,
        )
        self.last_button = NavigateButton(
            style=discord.ButtonStyle.grey,
            emoji="\N{BLACK RIGHT-POINTING DOUBLE TRIANGLE}",
            direction=self.source.get_max_pages,
            row=4,
            cog=cog,
        )
        self.close_button = CloseButton(
            style=discord.ButtonStyle.red,
            row=4,
            cog=cog,
        )
        self.refresh_button = RefreshButton(
            style=discord.ButtonStyle.grey,
            row=4,
            cog=cog,
        )
        self.select_view: PlaylistPlaySelector | PlaylistSelectSelector | None = None
        self.author = original_author

    @property
    def source(self) -> PlaylistPickerSource:
        return self._source

    async def prepare(self):
        self.clear_items()
        max_pages = self.source.get_max_pages()
        self.forward_button.disabled = False
        self.backward_button.disabled = False
        self.first_button.disabled = False
        self.last_button.disabled = False
        if max_pages == 1:
            self.forward_button.disabled = True
            self.backward_button.disabled = True
            self.first_button.disabled = True
            self.last_button.disabled = True
        elif max_pages == 2:
            self.first_button.disabled = True
            self.last_button.disabled = True
        self.add_item(self.close_button)
        self.add_item(self.first_button)
        self.add_item(self.backward_button)
        self.add_item(self.forward_button)
        self.add_item(self.last_button)
        if self.source.select_options:
            options = self.source.select_options
            self.remove_item(self.select_view)
            self.select_view = self.selector_cls(options, self.cog, self.selector_text, self.source.select_mapping)
            self.add_item(self.select_view)
        if self.select_view and not self.source.select_options:
            self.remove_item(self.select_view)
            self.select_view = None

    async def start(self, ctx: PyLavContext | DISCORD_INTERACTION_TYPE):
        if isinstance(ctx, discord.Interaction):
            ctx = await self.cog.bot.get_context(ctx)
        if ctx.interaction and not ctx.interaction.response.is_done():
            await ctx.defer(ephemeral=True)
        self.ctx = ctx
        await self.send_initial_message(ctx)

    async def show_page(self, page_number: int, interaction: DISCORD_INTERACTION_TYPE):
        await self._source.get_page(page_number)
        await self.prepare()
        self.current_page = page_number
        if self.message is not None:
            if not interaction.response.is_done():
                await interaction.response.pong()
            await self.message.edit(view=self)
        elif not interaction.response.is_done():
            await interaction.response.edit_message(view=self)

    async def wait_for_response(self):
        if isinstance(self.select_view, PlaylistSelectSelector):
            await asyncio.wait_for(self.select_view.responded.wait(), timeout=self.timeout)
            self.result = self.select_view.playlist


class PlaylistCreationFlow(discord.ui.View):
    ctx: DISCORD_CONTEXT_TYPE
    message: discord.Message
    url_prompt: PromptForInput
    name_prompt: PromptForInput
    scope_prompt: PromptForInput
    author: discord.abc.User

    def __init__(self, cog: DISCORD_COG_TYPE, original_author: discord.abc.User, *, timeout: int = 120) -> None:
        super().__init__(timeout=timeout)

        self.cog = cog
        self.bot = cog.bot
        self.author = original_author
        self.url_prompt = PromptForInput(
            cog=self.cog,
            title=shorten_string(max_length=100, string=_("Please enter the playlist URL.")),
            label=shorten_string(max_length=100, string=_("Playlist URL")),
            style=discord.TextStyle.paragraph,
            max_length=4000,
        )
        self.name_prompt = PromptForInput(
            cog=self.cog,
            title=shorten_string(max_length=100, string=_("Please enter the playlist name.")),
            label=shorten_string(max_length=100, string=_("Playlist Name")),
            max_length=64,
        )

        self.name_button = PlaylistUpsertButton(
            style=discord.ButtonStyle.grey,
            row=0,
            cog=cog,
            emoji=emojis.NAME,
            op="name",
        )
        self.url_button = PlaylistUpsertButton(
            style=discord.ButtonStyle.grey,
            row=0,
            cog=cog,
            emoji=emojis.URL,
            op="url",
        )
        self.done_button = DoneButton(
            style=discord.ButtonStyle.green,
            row=0,
            cog=cog,
        )
        self.queue_button = PlaylistQueueButton(
            style=discord.ButtonStyle.green,
            row=0,
            cog=cog,
            emoji=emojis.QUEUE,
        )
        self.close_button = CloseButton(
            style=discord.ButtonStyle.red,
            row=0,
            cog=cog,
        )

        self.name = None
        self.url = None
        self.scope = None
        self.done = False
        self.queue = None
        self.add_item(self.done_button)
        self.add_item(self.close_button)
        self.add_item(self.name_button)
        self.add_item(self.url_button)
        self.add_item(self.queue_button)

    async def send_initial_message(
        self, ctx: PyLavContext | DISCORD_INTERACTION_TYPE, description: str = None, title: str = None
    ):
        self.ctx = ctx
        self.message = await ctx.send(
            embed=await self.cog.pylav.construct_embed(description=description, title=title, messageable=ctx),
            view=self,
            ephemeral=True,
        )
        return self.message

    async def start(self, ctx: PyLavContext | DISCORD_INTERACTION_TYPE, description: str = None, title: str = None):
        await self.send_initial_message(ctx, description=description, title=title)

    async def interaction_check(self, interaction: DISCORD_INTERACTION_TYPE):
        """Just extends the default reaction_check to use owner_ids"""
        if (not await self.bot.allowed_by_whitelist_blacklist(interaction.user, guild=interaction.guild)) or (
            self.author and (interaction.user.id != self.author.id)
        ):
            await interaction.response.send_message(
                content=_("You are not authorized to interact with this."), ephemeral=True
            )
            return False
        return True

    async def prompt_url(self, interaction: DISCORD_INTERACTION_TYPE) -> None:
        await interaction.response.send_modal(self.url_prompt)
        await self.url_prompt.responded.wait()
        self.url = self.url_prompt.response

    async def prompt_name(self, interaction: DISCORD_INTERACTION_TYPE) -> None:
        await interaction.response.send_modal(self.name_prompt)
        await self.name_prompt.responded.wait()

        self.name = self.name_prompt.response

    async def prompt_scope(self, interaction: DISCORD_INTERACTION_TYPE) -> None:
        await interaction.response.send_modal(self.scope_prompt)
        await self.scope_prompt.responded.wait()
        self.scope = self.scope_prompt.response

    async def on_timeout(self):
        if self.message is None:
            return
        with contextlib.suppress(discord.HTTPException):
            if not self.message.flags.ephemeral:
                await self.message.delete()
            else:
                await self.message.edit(view=None)

    def stop(self):
        super().stop()
        asyncio.ensure_future(self.on_timeout())


class PlaylistManageFlow(discord.ui.View):
    ctx: DISCORD_CONTEXT_TYPE
    message: discord.Message
    url_prompt: PromptForInput
    name_prompt: PromptForInput
    scope_prompt: PromptForInput
    author: discord.abc.User

    def __init__(
        self,
        cog: DISCORD_COG_TYPE,
        original_author: discord.abc.User,
        playlist: Playlist,
        *,
        timeout: int = 120,
        manageable: bool = True,
    ) -> None:
        super().__init__(timeout=timeout)

        self.completed = asyncio.Event()
        self.cog = cog
        self.bot = cog.bot
        self.author = original_author
        self.playlist = playlist
        self.url_prompt = PromptForInput(
            cog=self.cog,
            title=shorten_string(max_length=100, string=_("Please enter the new URL for this playlist.")),
            label=shorten_string(max_length=100, string=_("Playlist URL")),
            style=discord.TextStyle.paragraph,
            max_length=4000,
        )
        self.name_prompt = PromptForInput(
            cog=self.cog,
            title=shorten_string(max_length=100, string=_("Please enter the new name for this playlist.")),
            label=shorten_string(max_length=100, string=_("Playlist Name")),
            max_length=64,
        )

        self.add_prompt = PromptForInput(
            cog=self.cog,
            title=_("What is the query to add to this playlist?"),
            label=_("Query"),
            style=discord.TextStyle.paragraph,
            max_length=4000,
        )

        self.remove_prompt = PromptForInput(
            cog=self.cog,
            title=_("What is the query to remove from this playlist?"),
            label=_("Query"),
            style=discord.TextStyle.paragraph,
            max_length=4000,
        )

        self.name_button = PlaylistUpsertButton(
            style=discord.ButtonStyle.grey,
            cog=cog,
            emoji=emojis.NAME,
            op="name",
        )
        self.url_button = PlaylistUpsertButton(
            style=discord.ButtonStyle.grey,
            cog=cog,
            emoji=emojis.URL,
            op="url",
        )
        self.add_button = PlaylistUpsertButton(
            style=discord.ButtonStyle.green,
            cog=cog,
            op="add",
            emoji=emojis.PLUS,
        )
        self.remove_button = PlaylistUpsertButton(
            style=discord.ButtonStyle.red,
            cog=cog,
            op="remove",
            emoji=emojis.MINUS,
        )

        self.done_button = DoneButton(
            style=discord.ButtonStyle.green,
            cog=cog,
        )
        self.delete_button = PlaylistDeleteButton(
            style=discord.ButtonStyle.red,
            cog=cog,
        )
        self.clear_button = PlaylistClearButton(
            style=discord.ButtonStyle.red,
            cog=cog,
        )
        self.close_button = CloseButton(
            style=discord.ButtonStyle.red,
            cog=cog,
        )
        self.update_button = PlaylistUpdateButton(
            style=discord.ButtonStyle.green,
            cog=cog,
        )

        self.download_button = PlaylistDownloadButton(
            style=discord.ButtonStyle.blurple,
            cog=cog,
            emoji=emojis.DOWNLOAD,
        )

        self.playlist_enqueue_button = EnqueuePlaylistButton(
            cog=cog,
            style=discord.ButtonStyle.blurple,
            emoji=emojis.PLAYLIST,
            playlist=playlist,
        )
        self.playlist_info_button = PlaylistInfoButton(
            cog=cog,
            style=discord.ButtonStyle.blurple,
            emoji=emojis.INFO,
            playlist=playlist,
        )
        self.queue_button = PlaylistQueueButton(
            style=discord.ButtonStyle.green,
            cog=cog,
            emoji=emojis.QUEUE,
        )
        self.dedupe_button = PlaylistDedupeButton(
            style=discord.ButtonStyle.red,
            cog=cog,
            emoji=emojis.DUPLICATE,
        )

        self.name = None
        self.url = None
        self.scope = None
        self.add_tracks = set()
        self.remove_tracks = set()

        self.clear = None
        self.delete = None
        self.cancelled = True
        self.done = False
        self.update = False
        self.queue = None
        self.dedupe = None
        if manageable or self.playlist.id in BUNDLED_PLAYLIST_IDS:
            self.add_item(self.done_button)
        self.add_item(self.close_button)
        if manageable:
            self.add_item(self.delete_button)
            self.add_item(self.clear_button)
        self.add_item(self.update_button)
        if manageable:
            self.add_item(self.name_button)
            self.add_item(self.url_button)
            self.add_item(self.add_button)
            self.add_item(self.remove_button)
        self.add_item(self.download_button)

        self.add_item(self.playlist_enqueue_button)
        self.add_item(self.playlist_info_button)
        if manageable:
            self.add_item(self.queue_button)
            self.add_item(self.dedupe_button)

    async def send_initial_message(
        self, ctx: PyLavContext | DISCORD_INTERACTION_TYPE, description: str = None, title: str = None
    ):
        self.ctx = ctx
        if not ctx.channel.permissions_for(ctx.me).attach_files:
            self.download_button.disabled = True
        self.message = await ctx.send(
            embed=await self.cog.pylav.construct_embed(description=description, title=title, messageable=ctx),
            view=self,
            ephemeral=True,
        )
        return self.message

    async def start(self, ctx: PyLavContext | DISCORD_INTERACTION_TYPE, description: str = None, title: str = None):
        await self.send_initial_message(ctx, description=description, title=title)

    async def interaction_check(self, interaction: DISCORD_INTERACTION_TYPE):
        """Just extends the default reaction_check to use owner_ids"""
        if (not await self.bot.allowed_by_whitelist_blacklist(interaction.user, guild=interaction.guild)) or (
            self.author and (interaction.user.id != self.author.id)
        ):
            await interaction.response.send_message(
                content=_("You are not authorized to interact with this."), ephemeral=True
            )
            return False
        return True

    async def prompt_url(self, interaction: DISCORD_INTERACTION_TYPE) -> None:
        self.cancelled = False
        await interaction.response.send_modal(self.url_prompt)
        await self.url_prompt.responded.wait()
        self.url = self.url_prompt.response

    async def prompt_name(self, interaction: DISCORD_INTERACTION_TYPE) -> None:
        self.cancelled = False
        await interaction.response.send_modal(self.name_prompt)
        await self.name_prompt.responded.wait()

        self.name = self.name_prompt.response

    async def prompt_scope(self, interaction: DISCORD_INTERACTION_TYPE) -> None:
        self.cancelled = False
        await interaction.response.send_modal(self.scope_prompt)
        await self.scope_prompt.responded.wait()
        self.scope = self.scope_prompt.response

    async def prompt_add_tracks(self, interaction: DISCORD_INTERACTION_TYPE) -> None:
        self.cancelled = False
        await interaction.response.send_modal(self.add_prompt)
        await self.add_prompt.responded.wait()
        self.add_tracks.add(self.add_prompt.response)

    async def prompt_remove_tracks(self, interaction: DISCORD_INTERACTION_TYPE) -> None:
        self.cancelled = False
        await interaction.response.send_modal(self.remove_prompt)
        await self.remove_prompt.responded.wait()
        self.remove_tracks.add(self.remove_prompt.response)

    async def on_timeout(self):
        self.completed.set()
        if self.message is None:
            return
        with contextlib.suppress(discord.HTTPException):
            if not self.message.flags.ephemeral:
                await self.message.delete()
            else:
                await self.message.edit(view=None)

    def stop(self):
        super().stop()
        asyncio.ensure_future(self.on_timeout())



from __future__ import annotations

from pathlib import Path

import discord
from redbot.core.i18n import Translator

from pylav.helpers import emojis
from pylav.type_hints.bot import DISCORD_COG_TYPE, DISCORD_INTERACTION_TYPE

_ = Translator("PyLav", Path(__file__))


class PreviousTrackButton(discord.ui.Button):
    def __init__(self, cog: DISCORD_COG_TYPE, style: discord.ButtonStyle, row: int = None):
        super().__init__(
            style=style,
            emoji=emojis.PREVIOUS,
            row=row,
        )
        self.cog = cog

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        if not interaction.response.is_done():
            await interaction.response.defer(ephemeral=True, thinking=True)
        context = await self.cog.bot.get_context(interaction)
        await self.cog.command_previous.callback(self.cog, context)
        await self.view.prepare()
        kwargs = await self.view.get_page(self.view.current_page)
        attachments = []
        if "file" in kwargs:
            attachments = [kwargs.pop("file")]
        elif "files" in kwargs:
            attachments = kwargs.pop("files")
        if attachments:
            kwargs["attachments"] = attachments
        await self.view.message.edit(view=self.view, **kwargs)


class StopTrackButton(discord.ui.Button):
    def __init__(self, cog: DISCORD_COG_TYPE, style: discord.ButtonStyle, row: int = None):
        super().__init__(
            style=style,
            emoji=emojis.STOP,
            row=row,
        )
        self.cog = cog

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        if not interaction.response.is_done():
            await interaction.response.defer(ephemeral=True, thinking=True)
        context = await self.cog.bot.get_context(interaction)
        await self.cog.command_stop.callback(self.cog, context)
        await self.view.prepare()
        kwargs = await self.view.get_page(self.view.current_page)
        attachments = []
        if "file" in kwargs:
            attachments = [kwargs.pop("file")]
        elif "files" in kwargs:
            attachments = kwargs.pop("files")
        if attachments:
            kwargs["attachments"] = attachments
        await self.view.message.edit(view=self.view, **kwargs)


class PauseTrackButton(discord.ui.Button):
    def __init__(self, cog: DISCORD_COG_TYPE, style: discord.ButtonStyle, row: int = None):
        super().__init__(
            style=style,
            emoji=emojis.PAUSE,
            row=row,
        )
        self.cog = cog

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        if not interaction.response.is_done():
            await interaction.response.defer(ephemeral=True, thinking=True)
        context = await self.cog.bot.get_context(interaction)
        await self.cog.command_pause.callback(self.cog, context)
        await self.view.prepare()
        kwargs = await self.view.get_page(self.view.current_page)
        attachments = []
        if "file" in kwargs:
            attachments = [kwargs.pop("file")]
        elif "files" in kwargs:
            attachments = kwargs.pop("files")
        if attachments:
            kwargs["attachments"] = attachments
        await self.view.message.edit(view=self.view, **kwargs)


class ResumeTrackButton(discord.ui.Button):
    def __init__(self, cog: DISCORD_COG_TYPE, style: discord.ButtonStyle, row: int = None):
        super().__init__(
            style=style,
            emoji=emojis.PLAY,
            row=row,
        )
        self.cog = cog

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        if not interaction.response.is_done():
            await interaction.response.defer(ephemeral=True, thinking=True)
        context = await self.cog.bot.get_context(interaction)
        await self.cog.command_resume.callback(self.cog, context)
        await self.view.prepare()
        kwargs = await self.view.get_page(self.view.current_page)
        attachments = []
        if "file" in kwargs:
            attachments = [kwargs.pop("file")]
        elif "files" in kwargs:
            attachments = kwargs.pop("files")
        if attachments:
            kwargs["attachments"] = attachments
        await self.view.message.edit(view=self.view, **kwargs)


class SkipTrackButton(discord.ui.Button):
    def __init__(self, cog: DISCORD_COG_TYPE, style: discord.ButtonStyle, row: int = None):
        super().__init__(
            style=style,
            emoji=emojis.NEXT,
            row=row,
        )
        self.cog = cog

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        if not interaction.response.is_done():
            await interaction.response.defer(ephemeral=True, thinking=True)
        context = await self.cog.bot.get_context(interaction)
        await self.cog.command_skip.callback(self.cog, context)
        await self.view.prepare()
        kwargs = await self.view.get_page(self.view.current_page)
        attachments = []
        if "file" in kwargs:
            attachments = [kwargs.pop("file")]
        elif "files" in kwargs:
            attachments = kwargs.pop("files")
        if attachments:
            kwargs["attachments"] = attachments
        await self.view.message.edit(view=self.view, **kwargs)


class IncreaseVolumeButton(discord.ui.Button):
    def __init__(self, cog: DISCORD_COG_TYPE, style: discord.ButtonStyle, row: int = None):
        super().__init__(
            style=style,
            emoji=emojis.VOLUME_UP,
            row=row,
        )
        self.cog = cog

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        if not interaction.response.is_done():
            await interaction.response.defer(ephemeral=True, thinking=True)
        context = await self.cog.bot.get_context(interaction)
        await self.cog.command_volume_change_by.callback(self.cog, context, change_by=5)
        await self.view.prepare()
        kwargs = await self.view.get_page(self.view.current_page)
        attachments = []
        if "file" in kwargs:
            attachments = [kwargs.pop("file")]
        elif "files" in kwargs:
            attachments = kwargs.pop("files")
        if attachments:
            kwargs["attachments"] = attachments
        await self.view.message.edit(view=self.view, **kwargs)


class DecreaseVolumeButton(discord.ui.Button):
    def __init__(self, cog: DISCORD_COG_TYPE, style: discord.ButtonStyle, row: int = None):
        super().__init__(
            style=style,
            emoji=emojis.VOLUME_DOWN,
            row=row,
        )
        self.cog = cog

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        if not interaction.response.is_done():
            await interaction.response.defer(ephemeral=True, thinking=True)
        context = await self.cog.bot.get_context(interaction)
        await self.cog.command_volume_change_by.callback(self.cog, context, change_by=-5)
        await self.view.prepare()
        kwargs = await self.view.get_page(self.view.current_page)
        attachments = []
        if "file" in kwargs:
            attachments = [kwargs.pop("file")]
        elif "files" in kwargs:
            attachments = kwargs.pop("files")
        if attachments:
            kwargs["attachments"] = attachments
        await self.view.message.edit(view=self.view, **kwargs)


class ToggleRepeatButton(discord.ui.Button):
    def __init__(self, cog: DISCORD_COG_TYPE, style: discord.ButtonStyle, row: int = None):
        super().__init__(
            style=style,
            emoji=emojis.LOOP,
            row=row,
        )
        self.cog = cog

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        if not interaction.response.is_done():
            await interaction.response.defer(ephemeral=True, thinking=True)
        context = await self.cog.bot.get_context(interaction)
        player = context.player
        if not player:
            return await context.send(
                embed=await self.cog.pylav.construct_embed(
                    description=_("I am not connected to any voice channel at the moment."), messageable=interaction
                ),
                ephemeral=True,
            )
        await self.cog.command_repeat.callback(self.cog, context, queue=await player.config.fetch_repeat_current())
        await self.view.prepare()
        kwargs = await self.view.get_page(self.view.current_page)
        attachments = []
        if "file" in kwargs:
            attachments = [kwargs.pop("file")]
        elif "files" in kwargs:
            attachments = kwargs.pop("files")
        if attachments:
            kwargs["attachments"] = attachments
        await self.view.message.edit(view=self.view, **kwargs)


class QueueHistoryButton(discord.ui.Button):
    def __init__(self, cog: DISCORD_COG_TYPE, style: discord.ButtonStyle, row: int = None):
        super().__init__(
            style=style,
            emoji=emojis.PLAYLIST,
            row=row,
        )
        self.cog = cog

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        if not interaction.response.is_done():
            await interaction.response.defer(ephemeral=True, thinking=True)
        context = await self.cog.bot.get_context(interaction)
        if not (__ := context.player):
            return await context.send(
                embed=await self.cog.pylav.construct_embed(
                    description=_("I am not connected to any voice channel at the moment."), messageable=interaction
                ),
                ephemeral=True,
            )
        from pylav.extension.red.ui.menus.queue import QueueMenu
        from pylav.extension.red.ui.sources.queue import QueueSource

        await QueueMenu(
            cog=self.cog,
            bot=self.cog.bot,
            source=QueueSource(guild_id=interaction.guild.id, cog=self.cog, history=True),
            original_author=interaction.user,
            history=True,
        ).start(ctx=context)


class ToggleRepeatQueueButton(discord.ui.Button):
    def __init__(self, cog: DISCORD_COG_TYPE, style: discord.ButtonStyle, row: int = None):
        super().__init__(
            style=style,
            emoji=emojis.REPEAT,
            row=row,
        )
        self.cog = cog

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        if not interaction.response.is_done():
            await interaction.response.defer(ephemeral=True, thinking=True)
        context = await self.cog.bot.get_context(interaction)
        player = context.player
        if not player:
            return await context.send(
                embed=await self.cog.pylav.construct_embed(
                    description=_("I am not connected to any voice channel at the moment."), messageable=interaction
                ),
                ephemeral=True,
            )
        repeat_queue = bool(await player.config.fetch_repeat_current())
        await self.cog.command_repeat.callback(self.cog, context, queue=repeat_queue)
        await self.view.prepare()
        kwargs = await self.view.get_page(self.view.current_page)
        attachments = []
        if "file" in kwargs:
            attachments = [kwargs.pop("file")]
        elif "files" in kwargs:
            attachments = kwargs.pop("files")
        if attachments:
            kwargs["attachments"] = attachments
        await self.view.message.edit(view=self.view, **kwargs)


class ShuffleButton(discord.ui.Button):
    def __init__(self, cog: DISCORD_COG_TYPE, style: discord.ButtonStyle, row: int = None):
        super().__init__(
            style=style,
            emoji=emojis.RANDOM,
            row=row,
        )
        self.cog = cog

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        if not interaction.response.is_done():
            await interaction.response.defer(ephemeral=True, thinking=True)
        context = await self.cog.bot.get_context(interaction)
        await self.cog.command_shuffle.callback(self.cog, context)
        await self.view.prepare()
        kwargs = await self.view.get_page(self.view.current_page)
        attachments = []
        if "file" in kwargs:
            attachments = [kwargs.pop("file")]
        elif "files" in kwargs:
            attachments = kwargs.pop("files")
        if attachments:
            kwargs["attachments"] = attachments
        await self.view.message.edit(view=self.view, **kwargs)


class DisconnectButton(discord.ui.Button):
    def __init__(self, cog: DISCORD_COG_TYPE, style: discord.ButtonStyle, row: int = None):
        super().__init__(
            style=style,
            emoji=emojis.POWER,
            row=row,
        )
        self.cog = cog

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        if not interaction.response.is_done():
            await interaction.response.defer(ephemeral=True, thinking=True)
        context = await self.cog.bot.get_context(interaction)
        await self.cog.command_disconnect.callback(self.cog, context)
        self.view.stop()
        await self.view.on_timeout()


class EmptyQueueButton(discord.ui.Button):
    def __init__(self, cog: DISCORD_COG_TYPE, style: discord.ButtonStyle, row: int = None):
        super().__init__(
            style=style,
            emoji=emojis.TRASH,
            row=row,
        )
        self.cog = cog

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        if not interaction.response.is_done():
            await interaction.response.defer(ephemeral=True, thinking=True)
        context = await self.cog.bot.get_context(interaction)
        player = context.player
        if not player.queue.size():
            return await context.send(
                embed=await self.cog.pylav.construct_embed(
                    description=_("There is nothing in the queue."), messageable=interaction
                ),
                ephemeral=True,
            )
        player.queue.clear()
        await context.send(
            embed=await self.cog.pylav.construct_embed(
                description=_("Removed tracks from the queue."), messageable=interaction
            ),
            ephemeral=True,
        )
        await self.view.prepare()
        kwargs = await self.view.get_page(self.view.current_page)
        attachments = []
        if "file" in kwargs:
            attachments = [kwargs.pop("file")]
        elif "files" in kwargs:
            attachments = kwargs.pop("files")
        if attachments:
            kwargs["attachments"] = attachments
        await self.view.message.edit(view=self.view, **kwargs)


class EnqueueButton(discord.ui.Button):
    def __init__(
        self,
        cog: DISCORD_COG_TYPE,
        style: discord.ButtonStyle,
        row: int = None,
    ):
        self.cog = cog
        super().__init__(
            style=style,
            emoji=emojis.PLUS,
            row=row,
        )

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        from pylav.extension.red.ui.modals.queue import EnqueueModal

        modal = EnqueueModal(self.cog, _("What do you want to enqueue?"))
        await interaction.response.send_modal(modal)
        await self.cog.bot.get_context(interaction)
        await self.view.prepare()
        kwargs = await self.view.get_page(self.view.current_page)
        attachments = []
        if "file" in kwargs:
            attachments = [kwargs.pop("file")]
        elif "files" in kwargs:
            attachments = kwargs.pop("files")
        if attachments:
            kwargs["attachments"] = attachments
        await self.view.message.edit(view=self.view, **kwargs)


class RemoveFromQueueButton(discord.ui.Button):
    def __init__(
        self,
        cog: DISCORD_COG_TYPE,
        style: discord.ButtonStyle,
        row: int = None,
    ):
        self.cog = cog
        super().__init__(
            style=style,
            emoji=emojis.MINUS,
            row=row,
        )

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        if not interaction.response.is_done():
            await interaction.response.defer(ephemeral=True, thinking=True)
        context = await self.cog.bot.get_context(interaction)

        from pylav.extension.red.ui.menus.queue import QueuePickerMenu
        from pylav.extension.red.ui.sources.queue import QueuePickerSource

        picker = QueuePickerMenu(
            bot=self.cog.bot,
            cog=self.cog,
            source=QueuePickerSource(guild_id=interaction.guild.id, cog=self.cog),
            delete_after_timeout=True,
            starting_page=0,
            menu_type="remove",
            original_author=interaction.user,
        )
        await picker.start(context)
        await self.view.prepare()
        kwargs = await self.view.get_page(self.view.current_page)
        attachments = []
        if "file" in kwargs:
            attachments = [kwargs.pop("file")]
        elif "files" in kwargs:
            attachments = kwargs.pop("files")
        if attachments:
            kwargs["attachments"] = attachments
        await self.view.message.edit(view=self.view, **kwargs)


class PlayNowFromQueueButton(discord.ui.Button):
    def __init__(
        self,
        cog: DISCORD_COG_TYPE,
        style: discord.ButtonStyle,
        row: int = None,
    ):
        self.cog = cog
        super().__init__(
            style=style,
            emoji=emojis.MUSICAL_NOTE,
            row=row,
        )

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        if not interaction.response.is_done():
            await interaction.response.defer(ephemeral=True, thinking=True)
        context = await self.cog.bot.get_context(interaction)

        from pylav.extension.red.ui.menus.queue import QueuePickerMenu
        from pylav.extension.red.ui.sources.queue import QueuePickerSource

        picker = QueuePickerMenu(
            bot=self.cog.bot,
            cog=self.cog,
            source=QueuePickerSource(guild_id=interaction.guild.id, cog=self.cog),
            delete_after_timeout=True,
            starting_page=0,
            menu_type="play",
            original_author=interaction.user,
        )
        await picker.start(context)
        await picker.wait()
        await self.view.prepare()
        kwargs = await self.view.get_page(self.view.current_page)
        attachments = []
        if "file" in kwargs:
            attachments = [kwargs.pop("file")]
        elif "files" in kwargs:
            attachments = kwargs.pop("files")
        if attachments:
            kwargs["attachments"] = attachments
        await self.view.message.edit(view=self.view, **kwargs)
